from pathlib import Path
from qtpy.QtCore import QPoint
from pytestqt.qtbot import QtBot
from himena.testing import WidgetTester
from himena_builtins.qt.ipynb import QIpynbEdit
from himena_builtins._io import default_text_reader

def test_ipynb_widget(qtbot: QtBot, sample_dir: Path):
    ipynb_widget = QIpynbEdit()
    ipynb_widget.show()
    qtbot.addWidget(ipynb_widget)
    with WidgetTester(ipynb_widget) as tester:
        model = default_text_reader(sample_dir / "ipynb.ipynb")
        tester.update_model(model)
        tester.cycle_model()
        output_widget = ipynb_widget._cell_widgets[0]._output_widget
        assert output_widget is not None
        output_widget._make_contextmenu(QPoint(100, 100))
        control = ipynb_widget.control_widget()
        control._insert_code()
        control._insert_md()
        control._delete_cell()
        control._clear_outputs()



from __future__ import annotations
from concurrent.futures import Future
from typing import TYPE_CHECKING
from enum import Enum
from qtpy import QtWidgets as QtW, QtCore, QtGui
from qtpy.QtCore import Qt
from superqt.utils import FunctionWorker, GeneratorWorker, WorkerBase

from himena.consts import DefaultFontFamily
from himena.qt._qprogress import QLabeledCircularProgressBar

if TYPE_CHECKING:
    from himena.qt._qmain_window import QMainWindow
    from himena.qt._qtab_widget import QTabWidget


class Anchor(Enum):
    """Anchor position"""

    top_left = "top_left"
    top_right = "top_right"
    bottom_left = "bottom_left"
    bottom_right = "bottom_right"


class _QOverlayBase(QtW.QDialog):
    """Overlay widget appears at the fixed position."""

    def __init__(self, main: QMainWindow):
        super().__init__(main._tab_widget, Qt.WindowType.SubWindow)
        self._widget = None

        _layout = QtW.QVBoxLayout(self)
        _layout.setContentsMargins(2, 2, 2, 2)
        _layout.setSpacing(0)

        main._tab_widget.resized.connect(self.alignToParent)
        self.setAnchor(Anchor.bottom_right)
        self.setVisible(False)

    def addWidget(self, widget: QtW.QWidget):
        """Set the central widget."""
        if self._widget is not None:
            self.removeWidget()
        self.layout().addWidget(widget)
        self.resize(widget.sizeHint())
        self._widget = widget
        self.alignToParent()

    def removeWidget(self):
        """Remove the central widget."""
        self._widget.setParent(None)
        self._widget = None
        self.resize(QtCore.QSize(0, 0))

    def setAnchor(self, anc: Anchor | str) -> None:
        """Set anchor position of the overlay widget."""
        self._anchor = Anchor(anc)
        return self.alignToParent()

    def show(self):
        """Show the overlay widget with animation."""
        super().show()
        self.alignToParent()
        return None

    def alignToParent(self):
        """Position widget at the bottom right edge of the parent."""
        if not self.isVisible():
            return
        qtabwidget = self.parentWidget()
        if not qtabwidget:
            return
        if self._anchor == Anchor.bottom_left:
            self.alignBottomLeft()
        elif self._anchor == Anchor.bottom_right:
            self.alignBottomRight()
        elif self._anchor == Anchor.top_left:
            self.alignTopLeft()
        elif self._anchor == Anchor.top_right:
            self.alignTopRight()
        else:
            raise RuntimeError

    def viewRect(self) -> QtCore.QRect:
        """Return the parent table rect."""
        parent = self.parentWidget()
        if widget := parent.widget_area(0):
            return widget.rect()
        elif widget := parent.widget(0):
            return widget.rect()
        return self.rect()

    def alignTopLeft(self, offset=(3, 3)):
        pos = self.viewRect().topLeft()
        pos.setX(pos.x() + offset[0])
        pos.setY(pos.y() + offset[1])
        self.move(pos)

    def alignTopRight(self, offset=(3, 3)):
        pos = self.viewRect().topRight()
        pos.setX(pos.x() - self.rect().width() - offset[0])
        pos.setY(pos.y() + offset[1])
        self.move(pos)

    def alignBottomLeft(self, offset=(3, 3)):
        pos = self.viewRect().bottomLeft()
        pos.setX(pos.x() + offset[0])
        pos.setY(pos.y() - self.rect().height() - offset[1])
        self.move(pos)

    def alignBottomRight(self, offset=(3, 3)):
        pos = self.viewRect().bottomRight()
        pos.setX(pos.x() - self.rect().width() - offset[0])
        pos.setY(pos.y() - self.rect().height() - offset[1])
        self.move(pos)

    # fmt: off
    if TYPE_CHECKING:
        def parentWidget(self) -> QTabWidget: ...
    # fmt: on


class QNotificationWidget(_QOverlayBase):
    """The overlay widget appears at the fixed position."""

    def __init__(self, main: QMainWindow, duration: int = 500):
        """
        The overlay widget appears at the fixed position.

        Parameters
        ----------
        parent : QMainWindow
            Parent table stack
        duration : int, default is 500
            Animation duration in msec.
        """
        super().__init__(main)

        size_grip = QtW.QSizeGrip(self)
        size_grip.setFixedHeight(8)
        self.layout().addWidget(
            size_grip, 0, Qt.AlignmentFlag.AlignTop | Qt.AlignmentFlag.AlignLeft
        )
        effect = QtW.QGraphicsOpacityEffect(self)
        effect.setOpacity(0.9)
        self.setGraphicsEffect(effect)
        self._effect = effect
        self.opacity_anim = QtCore.QPropertyAnimation(self._effect, b"opacity", self)
        self.geom_anim = QtCore.QPropertyAnimation(self, b"geometry", self)
        self._duration = duration
        self._timer: QtCore.QTimer | None = None
        self._close_btn = QtW.QPushButton("✕")
        self._close_btn.setFixedSize(15, 15)
        self._close_btn.setParent(
            self, self._close_btn.windowFlags() | Qt.WindowType.FramelessWindowHint
        )
        self._close_btn.clicked.connect(self._hide)

    def hideLater(self, sec: float = 5):
        """Hide overlay widget after a delay."""
        self._timer = QtCore.QTimer(self)
        self._timer.setInterval(int(sec * 1000))
        self._timer.setSingleShot(True)
        self._timer.timeout.connect(self._hide)
        self._timer.start()
        return None

    def _hide(self):
        self._close_btn.hide()
        if self.isVisible():
            self.setVisible(False)
            self._timer = None

    def slide_in(self):
        """Run animation that fades in the dialog with a slight slide up."""
        self.resize(280, 120)
        self.alignToParent()
        geom = self.geometry()
        self.geom_anim.setDuration(200)
        self.geom_anim.setStartValue(geom.translated(0, 20))
        self.geom_anim.setEndValue(geom)
        self.geom_anim.setEasingCurve(QtCore.QEasingCurve.Type.OutQuad)
        self.geom_anim.start()

    def moveEvent(self, a0):
        super().moveEvent(a0)
        self._align_close_btn()

    def show(self):
        """Show the overlay widget with animation."""
        super().show()
        self.slide_in()

    def hide(self) -> None:
        """Hide the overlay widget with animation."""
        self._close_btn.hide()
        self.opacity_anim.setDuration(self._duration)
        self.opacity_anim.setStartValue(0.9)
        self.opacity_anim.setEndValue(0)
        self.opacity_anim.start()

        @self.opacity_anim.finished.connect
        def _on_vanished():
            if self.isVisible():
                self.setVisible(False)
            self.opacity_anim.finished.disconnect()

    def show_and_hide_later(self, sec: float = 5):
        """Show the overlay widget with animation and hide after a delay."""
        self.show()
        self.hideLater(sec)

    def enterEvent(self, a0: QtCore.QEvent) -> None:
        self._enter_event()
        return super().enterEvent(a0)

    def _enter_event(self):
        if self._timer is not None:
            self._timer.stop()
        self._close_btn.show()

    def leaveEvent(self, a0: QtCore.QEvent) -> None:
        self._leave_event()
        return super().leaveEvent(a0)

    def _leave_event(self):
        if self._timer is not None:
            self._timer.start()
        if not self.rect().contains(self.mapFromGlobal(QtGui.QCursor.pos())):
            self._close_btn.hide()

    def _align_close_btn(self):
        pos_loc = self.rect().topRight() - QtCore.QPoint(
            self._close_btn.width() + 5, -5
        )
        self._close_btn.move(self.mapToGlobal(pos_loc))

    def resizeEvent(self, a0):
        super().resizeEvent(a0)
        self._align_close_btn()
        return None


class QJobStack(_QOverlayBase):
    job_finished = QtCore.Signal(QtW.QListWidgetItem)

    def __init__(self, main: QMainWindow):
        super().__init__(main)
        self._list_widget = QtW.QListWidget()
        self.addWidget(self._list_widget)
        self.setAnchor(Anchor.bottom_left)
        self.job_finished.connect(self._on_job_finished)

    def add_future(self, future: Future, desc: str, total: int = 0):
        item = QtW.QListWidgetItem()
        labeled_pbar = QLabeledCircularProgressBar(desc)
        pbar = labeled_pbar.pbar()
        pbar.setButtonState("square")
        pbar.setValue(-1)

        @pbar.abortRequested.connect
        def _aborting():
            pass  # TODO: not working yet

        if future.done():
            return None
        future.add_done_callback(lambda _: self.job_finished.emit(item))
        self._add_item_for_future(item, labeled_pbar)

    def add_worker(self, worker: WorkerBase, desc: str, total: int = 0):
        labeled_pbar = QLabeledCircularProgressBar(desc)
        pbar = labeled_pbar.pbar()
        pbar.setButtonState("square")
        if isinstance(worker, FunctionWorker):
            pbar.setValue(-1)

        elif isinstance(worker, GeneratorWorker):
            _nyield = 0

            @worker.yielded.connect
            def _increment(*_):
                nonlocal _nyield
                _nyield += 1
                if _nyield > total:
                    value = -1
                else:
                    value = _nyield / total * 100
                return pbar.setValue(value)

        else:
            raise TypeError(f"Unsupported worker type: {type(worker)}")

        @pbar.abortRequested.connect
        def _aborting():
            if not worker.abort_requested:
                pbar.infiniteRequested.emit(True)
                worker.quit()

        item = QtW.QListWidgetItem()
        worker.started.connect(lambda: self._add_item_for_future(item, labeled_pbar))
        worker.finished.connect(lambda: self._on_job_finished(item))

    def _add_item_for_future(self, item: QtW.QListWidgetItem, widget: QtW.QWidget):
        lw = self._list_widget
        lw.addItem(item)
        lw.setIndexWidget(lw.model().index(lw.count() - 1, 0), widget)
        self.adjustHeight()
        self.show()

    def _on_job_finished(self, item: QtW.QListWidgetItem):
        lw = self._list_widget
        lw.takeItem(lw.row(item))
        if lw.count() == 0:
            self.hide()
        else:
            self.adjustHeight()

    def adjustHeight(self):
        height = min(20 * min(2, self._list_widget.count()) + 6, 200)
        self._list_widget.setFixedHeight(height)
        self.setFixedHeight(height + 4)
        self.alignToParent()


class QWhatsThisWidget(_QOverlayBase):
    def __init__(self, parent: QTabWidget):
        super().__init__(parent)
        self._close_btn = QtW.QPushButton("✕")
        self._close_btn.setFixedSize(15, 15)
        self._close_btn.setParent(
            self, self._close_btn.windowFlags() | Qt.WindowType.FramelessWindowHint
        )
        self._close_btn.clicked.connect(self._hide)
        self.setAnchor(Anchor.top_right)
        self.setFixedSize(480, 360)

    def _hide(self):
        self._close_btn.hide()
        self.hide()
        return None

    def set_text(self, text: str, style: str = "plain"):
        text_widget = QtW.QTextEdit()
        text_widget.setFont(QtGui.QFont(DefaultFontFamily, 10))
        if style == "plain":
            text_widget.setText(text)
        elif style == "markdown":
            text_widget.setMarkdown(text)
        elif style == "html":
            text_widget.setHtml(text)
        else:
            raise ValueError(f"Unknown style: {style}")
        self.addWidget(text_widget)

    def enterEvent(self, a0: QtCore.QEvent) -> None:
        self._close_btn.show()
        pos_loc = self.rect().topRight() - QtCore.QPoint(
            self._close_btn.width() + 5, -5
        )
        self._close_btn.move(self.mapToGlobal(pos_loc))
        return super().enterEvent(a0)

    def leaveEvent(self, a0: QtCore.QEvent) -> None:
        if not self.rect().contains(self.mapFromGlobal(QtGui.QCursor.pos())):
            self._close_btn.hide()
        return super().leaveEvent(a0)



from __future__ import annotations

import importlib
import csv
import json
from pathlib import Path
from typing import TYPE_CHECKING, Any

import numpy as np
from himena.types import WidgetDataModel
from himena.standards.model_meta import ImageMeta, TableMeta, TextMeta, ImagePlaySetting
from himena.standards.roi import RoiListModel
from himena.consts import StandardType
from himena.workflow import Workflow

if TYPE_CHECKING:
    from openpyxl.worksheet.worksheet import Worksheet
    from himena.standards import plotting as hplt
    from himena._providers import ReaderStore

###################
##### Readers #########
################################


def _infer_encoding(file_path: Path) -> str | None:
    import chardet

    with file_path.open("rb") as f:
        detector = chardet.UniversalDetector()
        for line in f:
            detector.feed(line)
            if detector.done:
                break
        detector.close()
    encoding = detector.result["encoding"]
    if encoding == "ascii":
        encoding = "utf-8"  # ascii is a subset of utf-8
    return encoding


def _infer_separator(file_path: Path, encoding: str | None = None) -> str:
    _seps = ["\t", ";", ","]
    with file_path.open("r", encoding=encoding) as f:
        first_line = f.readline()

    _count_table = {sep: first_line.count(sep) for sep in _seps}
    _max_sep: str = max(_count_table, key=_count_table.get)
    if _count_table[_max_sep] == 0:
        return ","
    return _max_sep


def default_text_reader(file_path: Path) -> WidgetDataModel:
    """Read text file."""
    suffix = file_path.suffix.rstrip("~")
    if suffix in (".html", ".htm"):
        typ = StandardType.HTML
    elif suffix == ".json":
        typ = StandardType.JSON
    elif suffix == ".svg":
        typ = StandardType.SVG
    elif suffix == ".md":
        typ = StandardType.MARKDOWN
    elif suffix == ".ipynb":
        typ = StandardType.IPYNB
    else:
        typ = StandardType.TEXT
    encoding = _infer_encoding(file_path)
    value = file_path.read_text(encoding=encoding)
    return WidgetDataModel(
        value=value,
        type=typ,
        source=file_path,
        extension_default=suffix,
        metadata=TextMeta(encoding=encoding),
    )


def default_plain_text_reader(file_path: Path) -> WidgetDataModel:
    encoding = _infer_encoding(file_path)
    value = file_path.read_text(encoding=encoding)
    return WidgetDataModel(
        value=value,
        type=StandardType.TEXT,
        source=file_path,
        extension_default=file_path.suffix,
        metadata=TextMeta(encoding=encoding),
    )


def default_image_reader(file_path: Path) -> WidgetDataModel:
    """Read image file."""
    from PIL import Image, ImageSequence

    image = Image.open(file_path)
    if file_path.suffix == ".gif":
        arr = np.stack(
            [np.array(frame.convert("RGB")) for frame in ImageSequence.Iterator(image)],
            axis=0,
        )
        is_rgb = True
        play_setting = ImagePlaySetting(
            interval=image.info.get("duration", 100) / 1000,
            mode="once" if image.info.get("loop", 1) == 0 else "loop",
        )
        axes = ["t", "y", "x", "c"]
    else:
        arr = np.array(image)
        is_rgb = arr.ndim == 3 and arr.shape[2] in (3, 4)
        play_setting = None
        axes = ["y", "x", "c"] if is_rgb else ["y", "x"]

    return WidgetDataModel(
        value=arr,
        type=StandardType.IMAGE,
        extension_default=file_path.suffix,
        metadata=ImageMeta(
            axes=axes,
            is_rgb=is_rgb,
            interpolation="linear",
            play_setting=play_setting,
        ),
    )


def default_zip_reader(file_path: Path) -> WidgetDataModel:
    """Read zip file as a model stack."""
    import zipfile
    import tempfile
    from himena._providers import ReaderStore

    store = ReaderStore.instance()
    models = []
    with tempfile.TemporaryDirectory() as tmpdir, zipfile.ZipFile(file_path, "r") as z:
        tmpdir = Path(tmpdir)
        z.extractall(tmpdir)
        for each in tmpdir.glob("*"):
            model = store.run(each)
            model.title = each.name
            models.append(model)
    return WidgetDataModel(value=models, type=StandardType.MODELS)


def _read_txt_as_numpy(file_path: Path, delimiter: str | None = None):
    encoding = _infer_encoding(file_path)
    sep = delimiter or _infer_separator(file_path, encoding)
    try:
        if file_path.stat().st_size == 0:
            arr = np.array([[]], dtype=np.dtypes.StringDType())
        else:
            arr = np.loadtxt(
                file_path,
                dtype=np.dtypes.StringDType(),
                delimiter=sep,
                encoding=encoding,
            )
    except ValueError:
        # If the file has different number of columns in each row, np.loadtxt fails.
        with file_path.open("r", encoding=encoding) as f:
            reader = csv.reader(f, delimiter=sep)
            ncols = 0
            rows = []
            for row in reader:
                rows.append(row)
                ncols = max(ncols, len(row))
            arr = np.zeros((len(rows), ncols), dtype=np.dtypes.StringDType())
            for i, row in enumerate(rows):
                arr[i, : len(row)] = row

    return WidgetDataModel(
        value=arr,
        type=StandardType.TABLE,
        extension_default=file_path.suffix,
        metadata=TableMeta(separator=sep),
    )


def default_csv_reader(file_path: Path) -> WidgetDataModel:
    """Read CSV file."""
    return _read_txt_as_numpy(file_path)


def default_tsv_reader(file_path: Path) -> WidgetDataModel:
    """Read TSV file."""
    return _read_txt_as_numpy(file_path, "\t")


def default_plot_reader(file_path: Path) -> WidgetDataModel:
    """Write plot layout to a json file."""
    from himena.standards import plotting

    with open(file_path) as f:
        js = json.load(f)
        if not isinstance(js, dict):
            raise ValueError(f"Expected a dictionary, got {type(js)}.")
        if not (typ := js.pop("type")):
            raise ValueError("'type' field not found in the JSON file.")
        plot_layout = plotting.BaseLayoutModel.construct(typ, js)
    return WidgetDataModel(
        value=plot_layout,
        type=StandardType.PLOT,
        extension_default=".plot.json",
    )


def default_roi_reader(file_path: Path) -> WidgetDataModel:
    """Read image ROIs from a json file."""
    return WidgetDataModel(
        value=RoiListModel.model_validate_json(file_path.read_text()),
        type=StandardType.ROIS,
    )


def default_excel_reader(file_path: Path) -> WidgetDataModel:
    """Read Excel file."""
    import openpyxl

    wb = openpyxl.load_workbook(file_path, data_only=False)
    data = {}
    for sheet in wb.sheetnames:
        ws = wb[sheet]
        sheet_data = []
        for row in ws.iter_rows():
            row_input = []
            for cell in row:
                if cell.value is None:
                    row_input.append("")
                else:
                    row_input.append(str(cell.value))
            sheet_data.append(row_input)

        data[sheet] = np.asarray(sheet_data, dtype=np.dtypes.StringDType())

    return WidgetDataModel(
        value=data,
        type=StandardType.EXCEL,
        extension_default=file_path.suffix,
    )


def default_array_reader(file_path: Path) -> WidgetDataModel:
    """Read array file."""
    arr = np.load(file_path)
    return WidgetDataModel(value=arr, type=StandardType.ARRAY)


def default_pickle_reader(file_path: Path) -> WidgetDataModel:
    """Read pickle file."""
    import pickle

    with file_path.open("rb") as f:
        value = pickle.load(f)
    if isinstance(value, WidgetDataModel):
        # picke is created by himena.
        return value
    else:
        # pickle is created probably by other library. Just read as type "any".
        return WidgetDataModel(value=value, type=StandardType.ANY)


def fallback_reader(file_path: Path | list[Path]) -> WidgetDataModel:
    return WidgetDataModel(value=file_path, type=StandardType.READER_NOT_FOUND)


class DataFrameReader:
    _type = StandardType.DATAFRAME

    def __init__(self, module: str, method: str, kwargs: dict[str, Any]):
        self._module = module
        self._method = method
        self._kwargs = kwargs

    def as_plot_type(self) -> DataFrameReader:
        return DataFramePlotReader(self._module, self._method, self._kwargs)

    def __repr__(self):
        return f"{self.__class__.__name__}<{self._module}.{self._method}>"

    def __call__(self, file_path: Path) -> WidgetDataModel:
        mod = importlib.import_module(self._module)
        method = getattr(mod, self._method)
        df = method(file_path, **self._kwargs)
        return WidgetDataModel(value=df, type=self._type)


class DataFramePlotReader(DataFrameReader):
    _type = StandardType.DATAFRAME_PLOT


def _custom_key(fp: Path):
    if fp.is_symlink():
        group_type = "l"
    elif fp.is_dir():
        group_type = "d"
    else:
        group_type = "f"
    return f"{group_type}{fp.name}"


def default_file_list_reader(file_path: Path | list[Path]) -> WidgetDataModel:
    """Read list of files."""
    from himena._providers import ReaderStore

    value: list[WidgetDataModel] = []
    if isinstance(file_path, Path):
        # sort by file name, directories first
        _iterator = sorted(
            file_path.glob("*"),
            key=_custom_key,
        )
    else:
        _iterator = iter(file_path)
    store = ReaderStore.instance()
    for path in _iterator:
        model = WidgetDataModel(
            value=_make_lazy_reader(store, path),
            type=StandardType.LAZY,
            title=path.name,
        )
        value.append(model)
    return WidgetDataModel(value=value, type=StandardType.MODELS)


def default_workflow_reader(path: Path) -> WidgetDataModel:
    """Read workflow node."""
    value = Workflow.model_validate_json(path.read_bytes())
    return WidgetDataModel(value=value, type=StandardType.WORKFLOW)


def _make_lazy_reader(store: ReaderStore, path: Path):
    return lambda: store.run(path)._with_source(path)


###################
##### Writers #########
###############################


def default_text_writer(model: WidgetDataModel[str], path: Path) -> None:
    """Write text file."""
    if isinstance(meta := model.metadata, TextMeta):
        encoding = meta.encoding
    else:
        encoding = None
    return path.write_text(model.value, encoding=encoding)


def default_table_writer(model: WidgetDataModel[np.ndarray], path: Path) -> None:
    """Write table data to a text file."""
    delimiter = None
    suffix = path.suffix.rstrip("~")
    if isinstance(meta := model.metadata, TableMeta):
        delimiter = meta.separator
    if delimiter is None:
        if suffix == ".tsv":
            delimiter = "\t"
        else:
            delimiter = ","
    np.savetxt(path, model.value, fmt="%s", delimiter=delimiter)


def default_image_writer(model: WidgetDataModel[np.ndarray], path: Path) -> None:
    """Write image file."""
    from PIL import Image

    if path.suffix == ".gif":
        if model.value.ndim != 4 or model.value.shape[3] not in (3, 4):
            raise ValueError("Only 4D array with 3 or 4 channels can be saved as GIF.")
        frames = [
            Image.fromarray(model.value[i].astype(np.uint8))
            for i in range(model.value.shape[0])
        ]
        if (
            isinstance(meta := model.metadata, ImageMeta)
            and (ps := meta.play_setting) is not None
        ):
            duration = int(ps.interval * 1000)
            loop = 1 if ps.mode == "loop" else 0
        else:
            duration = 100
            loop = 1
        frames[0].save(
            path,
            save_all=True,
            append_images=frames[1:],
            loop=loop,
            duration=duration,
        )
    else:
        Image.fromarray(model.value).save(path)


def default_dict_writer(model: WidgetDataModel[dict[str, Any]], path: Path) -> None:
    """Write parameters to a json file."""
    path.write_text(json.dumps(model.value, default=_json_default))


def default_plot_writer(
    model: WidgetDataModel[hplt.BaseLayoutModel], path: Path
) -> None:
    """Write plot layout to a json file."""
    js = model.value.model_dump_typed()
    path.write_text(json.dumps(js, default=_json_default))


def default_excel_writer(
    model: WidgetDataModel[dict[str, np.ndarray]],
    path: Path,
) -> None:
    """Write Excel file."""
    import openpyxl

    wb = openpyxl.Workbook()
    if active_sheet := wb.active:
        wb.remove(active_sheet)
    for sheet_name, table in model.value.items():
        ws: Worksheet = wb.create_sheet(sheet_name)
        for r, row in enumerate(table):
            for c, cell_str in enumerate(row):
                cell_str: str
                if cell_str.startswith("="):
                    cell_data_type = "f"
                else:
                    try:
                        float(cell_str)
                        cell_data_type = "n"
                    except ValueError:
                        cell_data_type = "s"
                ws.cell(r + 1, c + 1).value = cell_str
                ws.cell(r + 1, c + 1).data_type = cell_data_type
    wb.save(path)


def default_array_writer(
    model: WidgetDataModel[np.ndarray],
    path: Path,
) -> None:
    """Write array file."""
    np.save(path, model.value)


def default_dataframe_writer(
    model: WidgetDataModel[dict[str, np.ndarray]],
    path: Path,
) -> None:
    """Write dataframe file."""
    from himena.data_wrappers import wrap_dataframe

    return wrap_dataframe(model.value).write(path)


def default_models_writer(
    model: WidgetDataModel[list[WidgetDataModel]],
    path: Path,
) -> None:
    """Write list of files."""
    from himena._providers import WriterStore

    store = WriterStore.instance()
    if path.suffix == "":
        if not path.exists():
            path.mkdir(parents=True)
        else:
            raise FileExistsError(f"Directory already exists: {path}")
        for each in model.value:
            file_name = _model_to_file_name(each)
            store.run(each, path / file_name)
    elif path.suffix == ".zip":
        import zipfile
        import tempfile

        with (
            tempfile.TemporaryDirectory() as tmpdir,
            zipfile.ZipFile(path, "w") as z,
        ):
            tmpdir = Path(tmpdir)
            for each in model.value:
                file_name = _model_to_file_name(each)
                store.run(each, tmpdir / file_name)
                z.write(tmpdir / file_name, arcname=file_name)
    else:
        raise ValueError(f"Unsupported file type: {path.suffix}")


def default_pickle_writer(model: WidgetDataModel[Any], path: Path) -> None:
    """Write pickle file."""
    import pickle

    with path.open("wb") as f:
        pickle.dump(model.value, f)


def default_roi_writer(model: WidgetDataModel[RoiListModel], path: Path) -> None:
    """Write image ROIs to a json file."""
    path.write_text(json.dumps(model.value.model_dump_typed(), default=_json_default))


def default_workflow_writer(model: WidgetDataModel[Workflow], path: Path) -> None:
    """Write workflow node."""
    path.write_text(model.value.model_dump_json())


def _json_default(obj):
    import cmap

    if isinstance(obj, np.ndarray):
        return obj.tolist()
    elif isinstance(obj, cmap.Color):
        return obj.hex
    elif isinstance(obj, cmap.Colormap):
        return obj.name
    raise TypeError(f"Object of type {type(obj).__name__} is not JSON serializable.")


def _model_to_file_name(m: WidgetDataModel):
    if not m.title:
        raise ValueError("Title is missing for one of the models.")
    if any(char in m.title for char in r'\/:*?"<>|'):
        raise ValueError(f"Invalid characters in file name: {m.title}")
    if Path(m.title).suffix == "":
        if m.extension_default:
            file_name = m.title + m.extension_default
        else:
            file_name = m.title
    else:
        file_name = m.title
    return file_name



from __future__ import annotations

from functools import wraps
from pathlib import Path
from typing import Any, Callable, ForwardRef, overload
from himena.types import WidgetDataModel
from himena.utils.misc import PluginInfo
from himena._providers import ReaderStore, WriterStore
from himena._utils import get_widget_data_model_type_arg


def _plugin_info_from_func(func: Callable) -> PluginInfo | None:
    if hasattr(func, "__module__"):
        module = func.__module__
        if module == "__main__" or "<" in module:
            # this plugin will never be available. Skip it.
            return None
        if hasattr(func, "__qualname__"):
            qual = func.__qualname__
            if not qual.isidentifier():
                return None
            return PluginInfo(module, qual)
        if hasattr(func, "__name__"):
            return PluginInfo(module, func.__name__)
    return None


class _IOPluginBase:
    __qualname__: str
    __module__: str
    __name__: str

    def __init__(
        self,
        func: Callable,
        matcher: Callable | None = None,
        *,
        priority: int = 100,
    ):
        self._priority = _check_priority(priority)
        self._func = func
        self._matcher = matcher or self._undefined_matcher
        self._plugin = _plugin_info_from_func(func)
        self.__name__ = str(func)  # default value
        wraps(func)(self)

    @property
    def priority(self) -> int:
        return self._priority

    @property
    def plugin(self) -> PluginInfo | None:
        return self._plugin

    @property
    def plugin_str(self) -> str | None:
        return self._plugin.to_str() if self._plugin else None

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.plugin_str or self.__name__}>"

    def _undefined_matcher(self, *_):
        raise NotImplementedError(
            f"Matcher for {self!r} is not defined. Use `define_matcher` to define a matcher for this plugin."
        )


class ReaderPlugin(_IOPluginBase):
    def __init__(
        self,
        reader: Callable[[Path | list[Path]], WidgetDataModel],
        matcher: Callable[[Path | list[Path]], bool] | None = None,
        *,
        priority: int = 100,
    ):
        super().__init__(reader, matcher, priority=priority)
        self._skip_if_list = False
        if hasattr(reader, "__annotations__"):
            annot_types = list(reader.__annotations__.values())
            if len(annot_types) == 1 and annot_types[0] in (
                Path,
                "Path",
                ForwardRef("Path"),
            ):
                self._skip_if_list = True

    def read(self, path: Path | list[Path]) -> WidgetDataModel:
        """Read file(s) and return a data model."""
        if isinstance(path, list):
            paths: list[Path] = []
            for p in path:
                if not p.exists():
                    raise FileNotFoundError(f"File {p!r} does not exist.")
                paths.append(p)
            out = self._func(paths)
        else:
            path = Path(path)
            if not path.exists():
                raise FileNotFoundError(f"File {path!r} does not exist.")
            out = self._func(path)
        if not isinstance(out, WidgetDataModel):
            raise TypeError(f"Reader plugin {self!r} did not return a WidgetDataModel.")
        return out

    __call__ = read

    def match_model_type(self, path: Path | list[Path]) -> str | None:
        """True if the reader can read the file."""
        if self._skip_if_list and isinstance(path, list):
            return None
        if self._matcher is None:
            return None
        out = self._matcher(path)
        if out is None or isinstance(out, str):
            return out
        raise TypeError(f"Matcher {self._matcher!r} did not return a string.")

    def define_matcher(self, matcher: Callable[[Path | list[Path]], str | None]):
        """Mark a function as a matcher.

        The matcher function should return a type string if the reader can read the
        file, or None otherwise. If the reader function is annotated with `Path`, only
        single Path input is forwarded to the matcher function, otherwise both `Path`
        and `list[Path]` will be considered.

        Examples
        --------
        A reader plugin that reads only text files:

        ```python
        @my_reader.define_matcher
        def _(path: Path):
            if path.suffix == ".txt":
                return "text"
            return None
        ```
        """
        # NOTE: matcher don't have to return the priority. If users want to define
        # a plugin that has different priority for different file type, they can just
        # split the plugin function into two.
        if self._matcher is self._undefined_matcher:
            raise ValueError(f"Matcher for {self!r} is already defined.")
        self._matcher = matcher
        return matcher

    def read_and_update_source(self, source: Path | list[Path]) -> WidgetDataModel:
        """Update workflow to a local-reader method if it is not set."""
        model = self.read(source)
        if len(model.workflow) == 0:
            model = model._with_source(source=source, plugin=self.plugin)
        return model


class WriterPlugin(_IOPluginBase):
    def __init__(
        self,
        writer: Callable[[WidgetDataModel, Path], Any],
        matcher: Callable[[Path | list[Path]], bool] | None = None,
        *,
        priority: int = 100,
    ):
        super().__init__(writer, matcher, priority=priority)
        if arg := get_widget_data_model_type_arg(writer):
            self._value_type_filter = arg
        else:
            self._value_type_filter = None

    def write(self, model: WidgetDataModel, path: Path) -> None:
        return self._func(model, path)

    __call__ = write

    def match_input(self, model: WidgetDataModel, path: Path) -> bool:
        if self._value_type_filter is not None and not isinstance(
            model.value, self._value_type_filter
        ):
            return False
        return self._matcher(model, path)

    def define_matcher(
        self, matcher: Callable[[WidgetDataModel, Path], bool]
    ) -> WriterPlugin:
        """Define how to match the input data model and the save path to this writer.

        Examples
        --------
        ```python
        @my_writer.define_matcher
        def _(model: WidgetDataModel, path: Path) -> bool:
            return path.suffix == ".txt" and model.type == "text"
        """
        self._matcher = matcher
        return self


@overload
def register_reader_plugin(
    reader: Callable[[Path | list[Path]], WidgetDataModel],
    *,
    priority: int = 100,
) -> ReaderPlugin: ...
@overload
def register_reader_plugin(
    *,
    priority: int = 100,
) -> Callable[[Callable[[Path | list[Path]], WidgetDataModel]], ReaderPlugin]: ...


def register_reader_plugin(reader=None, *, priority=100):
    """Register a reader plugin function.

    Decorate a function to register it as a reader plugin. The function should take a
    `Path` or a list of `Path`s as input and return a WidgetDataModel.

    ``` python
    from himena.plugins import register_reader_plugin

    @register_reader_plugin
    def my_reader(path) -> WidgetDataModel:
        ...  # read file and return a WidgetDataModel
    ```

    You will need to define a matcher function to tell whether this function can read
    a path using `define_matcher` method.

    ```python
    from himena import StandardType

    @my_reader.define_matcher
    def _(path: Path):
        if path.suffix == ".txt":
            return StandardType.TEXT  # StandardType.TEXT == "text"
        return None
    ```

    Parameters
    ----------
    priority : int, default 100
        Priority of choosing this reader when multiple readers are available. The
        default value 100 is higher than the himena builtin readers, so that your reader
        will prioritized over the default ones. If priority is less than 0, it will not
        be used unless users intentionally choose this plugin.
    """

    def _inner(func):
        if not callable(func):
            raise ValueError("Reader plugin must be callable.")
        ins = ReaderStore().instance()

        reader_plugin = ReaderPlugin(func, priority=priority)
        ins.add_reader(reader_plugin)
        return reader_plugin

    return _inner if reader is None else _inner(reader)


@overload
def register_writer_plugin(
    writer: Callable[[WidgetDataModel, Path], Any],
    *,
    priority: int = 100,
) -> WriterPlugin: ...
@overload
def register_writer_plugin(
    *,
    priority: int = 100,
) -> Callable[[Callable[[WidgetDataModel, Path], Any]], WriterPlugin]: ...


def register_writer_plugin(writer=None, *, priority=100):
    """Register a writer plugin function.

    Decorate a function to register it as a writer plugin. The function should take a
    `Path` as a save path and a `WidgetDataModel`.

    ``` python
    from himena.plugins import register_writer_plugin

    @register_writer_plugin
    def my_writer(path, model) -> None:
        ...  # read file and return a WidgetDataModel
    ```

    You will need to define a matcher function to tell whether this function can write
    a data model to the specified path using `define_matcher` method. Unlike reader
    plugins, matchers should return bool.

    ```python
    from himena import StandardType, WidgetDataModel

    @my_writer.define_matcher
    def _(path: Path, model: WidgetDataModel):
        if path.suffix == ".txt":
            return True
        return False
    ```

    Parameters
    ----------
    priority : int, default 100
        Priority of choosing this writer when multiple writers are available. The
        default value 100 is higher than the himena builtin writers, so that your writer
        will prioritized over the default ones. If priority is less than 0, it will not
        be used unless users intentionally choose this plugin.
    """

    def _inner(func):
        if not callable(func):
            raise ValueError("Writer plugin must be callable.")
        ins = WriterStore().instance()

        writer_plugin = WriterPlugin(func, priority=priority)
        ins.add_writer(writer_plugin)
        return writer_plugin

    return _inner if writer is None else _inner(writer)


def _check_priority(priority: int):
    if isinstance(priority, int) or hasattr(priority, "__int__"):
        return int(priority)
    raise TypeError(f"Priority must be an integer, not {type(priority)}.")



from __future__ import annotations

from concurrent.futures import Future, ThreadPoolExecutor
from contextlib import contextmanager
from logging import getLogger
from pathlib import Path
import threading
import uuid
from typing import (
    Any,
    Callable,
    Sequence,
    Generic,
    Iterator,
    Literal,
    TypeVar,
    overload,
    TYPE_CHECKING,
)
import warnings
from app_model.expressions import create_context
from psygnal import SignalGroup, Signal

from himena._app_model import AppContext, HimenaApplication
from himena._open_recent import RecentFileManager, RecentSessionManager
from himena._utils import get_widget_class_id, import_object
from himena.consts import NO_RECORDING_FIELD
from himena.plugins import _checker, actions as _actions
from himena.profile import AppProfile, load_app_profile
from himena.style import Theme
from himena.standards import BaseMetadata
from himena.types import (
    ClipboardDataModel,
    Size,
    WidgetDataModel,
    NewWidgetBehavior,
    DockArea,
    DockAreaString,
    BackendInstructions,
    WindowRect,
)
from himena.utils.misc import is_subtype
from himena.widgets._backend import BackendMainWindow
from himena.widgets._keyset import KeySet
from himena.widgets._hist import HistoryContainer, FileDialogHistoryDict
from himena.widgets._initialize import remove_instance
from himena.widgets._typemap import ObjectTypeMap, register_defaults
from himena.widgets._widget_list import TabList, TabArea, DockWidgetList
from himena.widgets._wrapper import ParametricWindow, SubWindow, DockWidget
from himena.workflow import ProgrammaticMethod, ActionHintRegistry
from himena._socket import SocketInfo

if TYPE_CHECKING:
    from app_model.types import KeyBindingRule
    from app_model.expressions import BoolOp
    from himena.widgets._widget_list import PathOrPaths

_W = TypeVar("_W")  # backend widget type
_T = TypeVar("_T")  # internal data type
_F = TypeVar("_F")  # function type
_LOGGER = getLogger(__name__)


class MainWindowEvents(SignalGroup, Generic[_W]):
    """Main window events."""

    tab_activated = Signal(TabArea[_W])
    window_activated = Signal(SubWindow[_W])
    window_added = Signal(SubWindow[_W])
    window_closed = Signal(SubWindow[_W])


class MainWindow(Generic[_W]):
    """The main window handler object."""

    def __init__(
        self,
        backend: BackendMainWindow[_W],
        app: HimenaApplication,  # must be initialized
        theme: Theme,
    ) -> None:
        from himena.widgets._initialize import set_current_instance

        self._events = MainWindowEvents()
        self._backend_main_window = backend
        self._internal_clipboard_data: Any | None = None
        self._tab_list = TabList(backend)
        self._new_widget_behavior = NewWidgetBehavior.WINDOW
        self._model_app = app
        self._instructions = BackendInstructions()
        self._history_tab = HistoryContainer[int](max_size=20)
        self._history_command = HistoryContainer[str](max_size=200)
        self._history_closed = HistoryContainer[tuple[Path, str | None]](max_size=10)
        self._file_dialog_hist = FileDialogHistoryDict()
        app.commands.executed.connect(self._on_command_execution)
        backend._connect_main_window_signals(self)
        self._ctx_keys = AppContext(create_context(self, max_depth=0))
        self._tab_list.changed.connect(backend._update_context)
        self._dock_widget_list = DockWidgetList(backend)
        self._recent_manager = RecentFileManager.default(app)
        self._recent_session_manager = RecentSessionManager.default(app)
        if "." not in app.name:
            # likely a mock instance
            set_current_instance(app.name, self)
            self._recent_manager.update_menu()
            self._recent_session_manager.update_menu()
        self._executor = ThreadPoolExecutor(max_workers=5)
        self._global_lock = threading.Lock()
        self._object_type_map = ObjectTypeMap()
        self.theme = theme
        register_defaults(self._object_type_map)

    @property
    def events(self) -> MainWindowEvents[_W]:
        """Main window events."""
        return self._events

    @property
    def object_type_map(self) -> ObjectTypeMap:
        """Mapping object to string that describes the type."""
        return self._object_type_map

    @property
    def action_hint_registry(self) -> ActionHintRegistry:
        """Action hint registry."""
        return _actions.AppActionRegistry.instance()._action_hint_reg

    @property
    def socket_info(self) -> SocketInfo:
        """Socket information."""
        eh = self._backend_main_window._event_loop_handler
        return SocketInfo(host=eh._host, port=eh._port)

    @property
    def keys(self) -> KeySet:
        """Get the set of currently pressed keys."""
        return KeySet(self)

    @property
    def theme(self) -> Theme:
        """Get the current color theme of the main window."""
        return self._theme

    @theme.setter
    def theme(self, theme: str | Theme) -> None:
        """Set the style of the main window."""
        if isinstance(theme, str):
            theme = Theme.from_global(theme)
        self._theme = theme
        self._backend_main_window._update_widget_theme(theme)

        # if child implements "theme_changed_callback", call it
        for win in self.iter_windows():
            _checker.call_theme_changed_callback(win.widget, theme)
        for dock in self.dock_widgets:
            _checker.call_theme_changed_callback(dock.widget, theme)
        return None

    @property
    def app_profile(self) -> AppProfile:
        """Get the current application profile object."""
        return load_app_profile(self._model_app.name)

    def submit_async_task(
        self,
        func: Callable,
        *args,
        progress_description: str | None = None,
        **kwargs,
    ) -> Future:
        """Submit a task to the thread pool.

        Parameters
        ----------
        func : callable
            Function to run in the background.
        progress_description : str, optional
            Description of the task in the progress bar.
        """
        future = self._executor.submit(func, *args, **kwargs)
        if progress_description is None:
            progress_description = f"Running {func!r}"
        self._backend_main_window._add_job_progress(
            future, desc=progress_description, total=0
        )
        self.model_app.injection_store.process(future)
        return future

    @property
    def tabs(self) -> TabList[_W]:
        """Tab list object."""
        return self._tab_list

    def windows_for_type(self, types: str | list[str]) -> list[SubWindow[_W]]:
        """Get all sub-windows for the given types."""
        windows = []
        if isinstance(types, str):
            types = [types]
        if tab := self.tabs.current():
            for win in tab:
                mtype = win.model_type()
                if mtype and any(is_subtype(mtype, t) for t in types):
                    windows.append(win)
        return windows

    @property
    def dock_widgets(self) -> DockWidgetList[_W]:
        """Dock widget list object."""
        return self._dock_widget_list

    @property
    def model_app(self) -> HimenaApplication:
        """The app-model application instance."""
        return self._model_app

    @property
    def area_size(self) -> tuple[int, int]:
        """(width, height) of the main window tab area."""
        return self._backend_main_window._area_size()

    @property
    def rect(self) -> WindowRect:
        """Window rect (left, top, width, height) of the main window."""
        return self._backend_main_window._main_window_rect()

    @rect.setter
    def rect(self, value) -> None:
        rect = WindowRect.from_tuple(*value)
        return self._backend_main_window._set_main_window_rect(rect)

    @property
    def size(self) -> Size:
        """Size (width, height) of the main window."""
        return self.rect.size()

    @size.setter
    def size(self, size) -> None:
        r0 = self.rect
        s0 = Size(*size)
        self.rect = WindowRect(r0.left, r0.top, s0.width, s0.height)

    @property
    def clipboard(self) -> ClipboardDataModel | None:
        """Get the clipboard data as a ClipboardDataModel instance."""
        model = self._backend_main_window._clipboard_data()
        model.internal_data = self._internal_clipboard_data
        return model

    @clipboard.setter
    def clipboard(self, data: str | ClipboardDataModel) -> None:
        """Set the clipboard data."""
        if isinstance(data, str):
            data = ClipboardDataModel(text=data)
        elif not isinstance(data, ClipboardDataModel):
            raise ValueError("Clipboard data must be a ClipboardDataModel instance.")
        _LOGGER.info("Setting clipboard data: %r", data)
        self._backend_main_window._set_clipboard_data(data)
        self._internal_clipboard_data = data.internal_data

    def set_clipboard(
        self,
        *,
        text: str | None = None,
        html: str | None = None,
        image: Any | None = None,
        files: list[str | Path] | None = None,
        internal_data: Any | None = None,
    ) -> None:
        """Set clipboard data."""
        self.clipboard = ClipboardDataModel(
            text=text,
            html=html,
            image=image,
            files=files or [],
            internal_data=internal_data,
        )

    def add_tab(self, title: str | None = None) -> TabArea[_W]:
        """Add a new tab of given name."""
        return self.tabs.add(title)

    def window_for_id(self, identifier: uuid.UUID) -> SubWindow[_W] | None:
        """Retrieve a sub-window by its identifier."""
        if not isinstance(identifier, uuid.UUID):
            raise ValueError(f"Expected UUID, got {identifier!r}.")
        for win in self.iter_windows():
            if win._identifier == identifier:
                return win

    def _window_for_workflow_id(self, identifier: uuid.UUID) -> SubWindow[_W] | None:
        """Retrieve a sub-window by its workflow identifier."""
        for win in self.iter_windows():
            try:
                last_id = win._widget_workflow.last_id()
            except Exception:
                return None
            if last_id == identifier:
                return win

    def _current_or_new_tab(self) -> tuple[int, TabArea[_W]]:
        if self._new_widget_behavior is NewWidgetBehavior.WINDOW:
            if len(self.tabs) == 0:
                self.add_tab()
                idx = 0
            else:
                idx = self._backend_main_window._current_tab_index()
            tabarea = self.tabs[idx]
        else:
            tabarea = self.add_tab()
            idx = len(self.tabs) - 1
        return idx, tabarea

    def add_widget(
        self,
        widget: _W,
        *,
        title: str | None = None,
    ) -> SubWindow[_W]:
        """Add a widget to the sub window.

        Any widget that can be interpreted by the backend can be added. For example, for
        Qt application, you can add any QWidget instance:

        ```python
        ui.add_widget(QtW.QLabel("Hello world!"), title="my widget!")
        ```

        Parameters
        ----------
        widget : Any
            Widget to add.
        title : str, optional
            Title of the sub-window. If not given, its name will be automatically
            generated.

        Returns
        -------
        SubWindow
            The sub-window handler.
        """
        _, tabarea = self._current_or_new_tab()
        return tabarea.add_widget(widget, title=title)

    def add_dock_widget(
        self,
        widget: _W,
        *,
        title: str | None = None,
        area: DockAreaString | DockArea | None = DockArea.RIGHT,
        allowed_areas: list[DockAreaString | DockArea] | None = None,
        _identifier: uuid.UUID | None = None,
    ) -> DockWidget[_W]:
        """Add a custom widget as a dock widget of the main window.

        Parameters
        ----------
        widget : Widget type
            Widget instance that is allowed for the backend.
        title : str, optional
            Title of the dock widget.
        area : dock area, default DockArea.RIGHT
            String or DockArea enum that describes where the dock widget initially
            appears.
        allowed_areas : list of dock area, optional
            List of allowed dock areas for the widget.

        Returns
        -------
        DockWidget
            The dock widget handler.
        """
        dock = DockWidget(widget, self._backend_main_window, identifier=_identifier)
        dock_native = dock._split_interface_and_frontend()[1]
        self._backend_main_window.add_dock_widget(
            dock_native, title=title, area=area, allowed_areas=allowed_areas
        )
        self._dock_widget_list._add_dock_widget(dock)
        _checker.call_widget_added_callback(widget)
        _checker.call_theme_changed_callback(widget, self.theme)
        return dock

    def add_object(
        self,
        value: Any,
        *,
        type: str | None = None,
        title: str | None = None,
        force_open_with: str | None = None,
        metadata: Any | None = None,
    ) -> SubWindow[_W]:
        """Add any data as a widget data model.

        CAUTION: result of this method may not be stable between different sessions. If
        type is not given, the application will look for the proper widget data type in
        the type map. If this type map was modified, the result will be different.

        Parameters
        ----------
        value : Any
            Any object. Whether it can be represented as a widget is dependent on the
            plugins that are installed.
        type : str, optional
            Any str that describes the type of the object. This type must be registered
            with a proper widget class.
        title : str, optional
            Title of the sub-window.

        Returns
        -------
        SubWindow
            The sub-window handler.
        """
        if type is None:
            if isinstance(metadata, BaseMetadata):
                type = metadata.expected_type()
        if type is None:
            type, value, metadata_fallback = self._object_type_map.pick_type(value)
            if metadata is None:
                metadata = metadata_fallback
        wd = WidgetDataModel(
            value=value,
            type=type,
            title=title,
            force_open_with=force_open_with,
            metadata=metadata,
            workflow=ProgrammaticMethod(output_model_type=type).construct_workflow(),
        )
        return self.add_data_model(wd)

    def add_data_model(self, model_data: WidgetDataModel) -> SubWindow[_W]:
        """Add a widget data model as a widget."""
        _, tabarea = self._current_or_new_tab()
        return tabarea.add_data_model(model_data)

    def add_function(
        self,
        func: Callable[..., _T],
        *,
        preview: bool = False,
        title: str | None = None,
        show_parameter_labels: bool = True,
        auto_close: bool = True,
        run_async: bool = False,
        result_as: Literal["window", "below", "right"] = "window",
    ) -> ParametricWindow[_W]:
        """Add a function as a parametric sub-window.

        The input function must return a `WidgetDataModel` instance, which can be
        interpreted by the application.

        Parameters
        ----------
        func : function (...) -> WidgetDataModel
            Function that generates a model from the input parameters.
        title : str, optional
            Title of the sub-window.

        Returns
        -------
        SubWindow
            The sub-window instance that represents the output model.
        """
        _, tabarea = self._current_or_new_tab()
        return tabarea.add_function(
            func, title=title, preview=preview, result_as=result_as,
            show_parameter_labels=show_parameter_labels, auto_close=auto_close,
            run_async=run_async,
        )  # fmt: skip

    def add_parametric_widget(
        self,
        widget: _W,
        callback: Callable | None = None,
        *,
        title: str | None = None,
        preview: bool = False,
        auto_close: bool = True,
        auto_size: bool = True,
        run_async: bool = False,
        result_as: Literal["window", "below", "right"] = "window",
    ) -> ParametricWindow[_W]:
        _, area = self._current_or_new_tab()
        return area.add_parametric_widget(
            widget, callback, title=title, preview=preview, auto_close=auto_close,
            auto_size=auto_size, result_as=result_as, run_async=run_async,
        )  # fmt: skip

    def read_file(
        self,
        file_path: PathOrPaths,
        plugin: str | None = None,
    ) -> SubWindow[_W]:
        """Read local file(s) and open as a new sub-window."""
        _, tabarea = self._current_or_new_tab()
        return tabarea.read_file(file_path, plugin=plugin)

    def read_files(self, file_paths: PathOrPaths):
        """Read multiple files one by one and open as new sub-windows in a same tab."""
        _, tabarea = self._current_or_new_tab()
        return tabarea.read_files(file_paths)

    def read_files_async(self, file_paths: PathOrPaths, plugin: str | None = None):
        """Read multiple files asynchronously and open as new sub-windows."""
        _, tabarea = self._current_or_new_tab()
        return tabarea.read_files_async(file_paths, plugin=plugin)

    def load_session(self, path: str | Path) -> None:
        """Read a session file and update the main window based on the content."""
        from himena.session import update_from_zip, update_from_directory

        fp = Path(path)
        if fp.suffix == ".zip":
            update_from_zip(self, fp)
        elif fp.is_dir():
            update_from_directory(self, fp)
        else:
            raise ValueError(f"Session must be a zip file or a directory, got {fp}.")
        # always plugin=None for reading a session file as a session
        self._recent_session_manager.append_recent_files([(fp, None)])
        self.set_status_tip(f"Session loaded: {fp}", duration=5)

    def save_session(
        self,
        path: str | Path,
        *,
        save_copies: bool = False,
        allow_calculate: Sequence[str] = (),
    ) -> None:
        """Save the current session to a zip file as a stand-along file."""
        from himena.session import dump_zip, dump_directory

        path = Path(path)
        if path.suffix == ".zip":
            dump_zip(
                self, path, save_copies=save_copies, allow_calculate=allow_calculate
            )
        else:
            dump_directory(
                self, path, save_copies=save_copies, allow_calculate=allow_calculate
            )
        self.set_status_tip(f"Session saved to {path}")
        self._recent_session_manager.append_recent_files([(path, None)])

    def clear(self) -> None:
        """Clear all widgets in the main window."""
        self.tabs.clear()
        self.dock_widgets.clear()

    def set_status_tip(
        self,
        text: str,
        duration: float = 10.0,
        process_event: bool = False,
    ) -> None:
        """Set the status tip of the main window.

        This method can be safely called from any thread.

        Parameters
        ----------
        text : str
            Text to show in the status bar.
        duration : float, default 10.0
            Duration (seconds) to show the status tip.
        process_event : bool, default False
            If True, the application will process events after setting the status tip.
        """
        self._backend_main_window._set_status_tip(text, duration)
        if process_event:
            self._backend_main_window._event_loop_handler.process_events()

    def show_notification(self, text: str, duration: float = 5.0) -> None:
        """Show a temporary notification in the main window.

        Parameters
        ----------
        text : str
            Text to show in the notification.
        duration : float, default 5.0
            Duration (seconds) to show the notification.
        """
        self._backend_main_window._show_notification(text, duration)

    def show_tooltip(
        self,
        text: str,
        duration: float = 3.0,
        behavior: Literal["stay", "follow", "until_move"] = "follow",
    ) -> None:
        """Show a temporary tooltip next to the cursor in the main window.

        Parameters
        ----------
        text : str
            HTML text to show in the tooltip.
        duration : float, default 3.0
            Duration (seconds) to show the tooltip.
        behavior : str, default "follow"
            Behavior of the tooltip. "stay" to show at the position where it is created,
            "follow" to follow the cursor, "until_move" to show until the cursor moves.
        """
        self._backend_main_window._show_tooltip(text, duration, behavior)

    @overload
    def register_function(
        self,
        func: None = None,
        *,
        menus: str | Sequence[str] = "plugins",
        title: str | None = None,
        types: str | Sequence[str] | None = None,
        enablement: BoolOp | None = None,
        keybindings: Sequence[KeyBindingRule] | None = None,
        command_id: str | None = None,
    ) -> None: ...  # noqa: E501
    @overload
    def register_function(
        self,
        func: _F,
        *,
        menus: str | Sequence[str] = "plugins",
        title: str | None = None,
        types: str | Sequence[str] | None = None,
        enablement: BoolOp | None = None,
        keybindings: Sequence[KeyBindingRule] | None = None,
        command_id: str | None = None,
    ) -> _F: ...  # noqa: E501

    def register_function(
        self,
        func=None,
        *,
        menus="plugins",
        title=None,
        types=None,
        enablement=None,
        keybindings=None,
        command_id=None,
    ):
        """Register a function as a callback in runtime.

        Example
        -------
        ``` python
        @ui.register_function(menus="plugins", title="Test functions)
        def test_function():
            print("test")
        ```

        Parameters
        ----------
        func : callable, optional
            Function to register as an action.
        menus : str or sequence of str, default "plugins"
            Menu(s) to add the action. Submenus are separated by `/`.
        title : str, optional
            Title of the action. Name of the function will be used if not given.
        types: str or sequence of str, optional
            The `type` parameter(s) allowed as the WidgetDataModel. If this parameter
            is given, action will be grayed out if the active window does not satisfy
            the listed types.
        enablement: Expr, optional
            Expression that describes when the action will be enabled. As this argument
            is a generalized version of `types` argument, you cannot use both of them.
        command_id : str, optional
            Command ID. If not given, the function qualname will be used.
        """

        def _inner(f):
            action = _actions.make_action_for_function(
                f,
                menus=menus,
                title=title,
                types=types,
                enablement=enablement,
                keybindings=keybindings,
                command_id=command_id,
            )
            _actions.AppActionRegistry.instance().add_action(action)
            added_menus = _actions.AppActionRegistry.instance().install_to(
                self.model_app, [action]
            )
            self._backend_main_window._rebuild_for_runtime(added_menus)
            return f

        return _inner(func) if func else _inner

    def exec_action(
        self,
        id: str,
        *,
        model_context: WidgetDataModel | None = None,
        window_context: SubWindow | None = None,
        with_params: dict[str, Any] | None = None,
        with_defaults: dict[str, Any] | None = None,
        process_model_output: bool = True,
    ) -> Any:
        """Execute an action by its ID.

        Parameters
        ----------
        id : str
            Action ID.
        model_context : WidgetDataModel, optional
            If given, this model will override the application context for the type
            `WidgetDataModel` before the execution.
        window_context : SubWindow, optional
            If given, this window will override the application context for the type
            `SubWindow` before the execution.
        with_params : dict, optional
            Parameters to pass to the parametric action. These parameters will directly
            be passed to the parametric window created after the action is executed.
        with_defaults : dict, optional
            If given, the resulting parametric window will be updated with these values.
        process_model_output : bool, default True
            If True, the output result will be processed by the application context. If
            the command return a `WidgetDataModel` instance, it will be converted to a
            sub-window.
        """
        if with_params is not None and with_defaults is not None:
            raise TypeError(
                "Cannot use both `with_params` and `with_defaults` at the same time."
            )
        providers: list[tuple[Any, type]] = []
        if model_context is not None:
            providers.append((model_context, WidgetDataModel, 1000))
        if window_context is not None:
            if isinstance(window_context, SubWindow):
                _window_context = window_context
            else:
                raise TypeError(
                    f"`window_context` must be SubWindow or UUID, got {window_context}"
                )
            providers.append((_window_context, SubWindow, 1000))
            if model_context is None and _window_context.supports_to_model:
                providers.append((_window_context.to_model(), WidgetDataModel, 100))
        # execute the command under the given context
        with (
            self.model_app.injection_store.register(providers=providers),
            self._execute_in_context(
                is_gui=with_params is None, process_model_output=process_model_output
            ),
        ):
            result = self.model_app.commands.execute_command(id).result()
            if with_params is not None:
                if (tab := self.tabs.current()) is not None and len(tab) > 0:
                    param_widget = tab[-1]
                else:  # pragma: no cover
                    raise RuntimeError("Unreachable code.")
                if not isinstance(param_widget, ParametricWindow):
                    if len(with_params) == 0:
                        if isinstance(result, Future):
                            return result.result()  # or appropriate handling
                        return result
                    raise ValueError(
                        f"Parametric widget expected but got {param_widget}."
                    )
                # run the callback with the given parameters synchronously
                result = param_widget._callback_with_params(
                    with_params,
                    force_sync=True,
                    force_close=True,
                )
            elif with_defaults is not None:
                if (tab := self.tabs.current()) is not None and len(tab) > 0:
                    param_widget = tab[-1]
                else:  # pragma: no cover
                    raise ValueError(
                        f"Command {id!r} did not create a parametric window."
                    )
                if not isinstance(param_widget, ParametricWindow):
                    if with_defaults:
                        raise ValueError(
                            f"Parametric widget expected but got {param_widget}."
                        )
                else:
                    param_widget.update_params(with_defaults)
        return result

    @overload
    def exec_choose_one_dialog(
        self,
        title: str,
        message: str,
        choices: list[tuple[str, _T]],
        how: Literal["buttons", "radiobuttons"] = "buttons",
    ) -> _T | None: ...
    @overload
    def exec_choose_one_dialog(
        self,
        title: str,
        message: str,
        choices: list[str],
        how: Literal["buttons", "radiobuttons"] = "buttons",
    ) -> str | None: ...

    def exec_choose_one_dialog(self, title, message, choices, how="buttons"):
        """Execute a dialog to choose one from the given choices.

        Parameters
        ----------
        title : str
            Window title of the dialog.
        message : str
            HTML Message to show in the dialog.
        choices : list
            List of choices. Each choice can be a string or a tuple of (text, value).
            This method will return the selected value.
        how : str, default "buttons"
            How to show the choices. "buttons" for horizontal buttons, "radiobuttons"
            for vertically arranged radio buttons.
        """
        if res := self._instructions.choose_one_dialog_response:
            return res()
        _choices_normed = []
        for choice in choices:
            if isinstance(choice, str):
                _choices_normed.append((choice, choice))
            else:
                text, value = choice
                _choices_normed.append((text, value))
        return self._backend_main_window._request_choice_dialog(
            title, message, _choices_normed, how=how
        )

    @overload
    def exec_file_dialog(
        self,
        mode: Literal["r", "d", "w"] = "r",
        *,
        extension_default: str | None = None,
        allowed_extensions: list[str] | None = None,
        caption: str | None = None,
        start_path: str | Path | None = None,
        group: str | None = None,
    ) -> Path | None: ...
    @overload
    def exec_file_dialog(
        self,
        mode: Literal["rm"],
        *,
        extension_default: str | None = None,
        allowed_extensions: list[str] | None = None,
        caption: str | None = None,
        start_path: str | Path | None = None,
        group: str | None = None,
    ) -> list[Path] | None: ...

    def exec_file_dialog(
        self,
        mode: Literal["r", "d", "w", "rm"] = "r",
        *,
        extension_default=None,
        allowed_extensions=None,
        caption=None,
        start_path=None,
        group: str | None = None,
    ):
        """Execute a file dialog to get file path(s)."""
        if mode not in {"r", "d", "w", "rm"}:
            raise ValueError(f"`mode` must be 'r', 'd', 'w' or 'rm', got {mode!r}.")
        if res := self._instructions.file_dialog_response:
            return res()
        if group is None:
            group = mode

        if mode == "w":
            if start_path is None:
                _start_path = self._file_dialog_hist.get_path(group)
            elif Path(start_path).parent != Path("."):
                _start_path = Path(start_path)
            else:  # filename only is given
                _start_path = self._file_dialog_hist.get_path(group, str(start_path))
        else:
            _start_path = Path(start_path or self._file_dialog_hist.get_path(group))
        result = self._backend_main_window._open_file_dialog(
            mode,
            extension_default=extension_default,
            allowed_extensions=allowed_extensions,
            caption=caption,
            start_path=_start_path,
        )
        if result is None:
            return None
        if mode in ["r", "w", "d"]:
            self._file_dialog_hist.update(group, result.parent)
        elif result:
            self._file_dialog_hist.update(group, result[0].parent)
        return result

    def show(self, run: bool = False) -> None:
        """
        Show the main window.

        Parameters
        ----------
        run : bool, default False
            If True, run the application event loop.
        """
        self._backend_main_window.show(run)
        return None

    def close(self) -> None:
        """Close the main window."""
        self._backend_main_window._exit_main_window(confirm=False)
        remove_instance(self.model_app.name, self)
        return None

    @property
    def current_window(self) -> SubWindow[_W] | None:
        """Get the current sub-window."""
        idx_tab = self._backend_main_window._current_tab_index()
        if idx_tab is None:
            return None
        idx_win = self._backend_main_window._current_sub_window_index(idx_tab)
        if idx_win is None:
            return None
        return self.tabs[idx_tab][idx_win]

    @current_window.setter
    def current_window(self, win: SubWindow[_W] | None) -> None:
        """Set the current sub-window."""
        _main = self._backend_main_window
        if win is None:
            _main._set_current_tab_index(None)
            i_tab = _main._current_tab_index()
            if i_tab is not None:
                _main._set_current_sub_window_index(i_tab, None)
        else:
            for i_tab, tab in self.tabs.enumerate():
                for i_win, sub in tab.enumerate():
                    if sub is win:
                        _main._set_current_tab_index(i_tab)
                        _main._set_current_sub_window_index(i_tab, i_win)
                        return None
        return None

    @property
    def current_model(self) -> WidgetDataModel | None:
        """Get the current model of the active sub-window."""
        if sub := self.current_window:
            return sub.to_model()
        return None

    def iter_windows(self) -> Iterator[SubWindow[_W]]:
        """Iterate over all the sub-windows in this main window."""
        for tab in self.tabs:
            yield from tab

    def _provide_file_output(self) -> tuple[WidgetDataModel, SubWindow[_W]]:
        if sub := self.current_window:
            model = sub.to_model()
            return model, sub
        else:
            raise ValueError("No active window.")

    def _tab_activated(self, i: int):
        if i < 0:
            return None
        tab = self.tabs.get(i)
        if tab is not None:
            self.events.tab_activated.emit(tab)
            self._main_window_resized(self.area_size)  # update layout and anchor
        if self._history_tab.get_from_last(1) != i:
            self._history_tab.add(i)
        return None

    def move_window(self, sub: SubWindow[_W], target_index: int) -> None:
        """Move the sub-window to the target tab index."""
        i_tab = i_win = None
        for _i_tab, tab in self.tabs.enumerate():
            for _i_win, win in tab.enumerate():
                if win is sub:
                    i_tab = _i_tab
                    i_win = _i_win
                    break

        if i_tab is None or i_win is None or target_index == i_tab:
            return None
        title = self.tabs[i_tab][i_win].title
        old_rect = self.tabs[i_tab][i_win].rect
        win, widget = self.tabs[i_tab]._pop_no_emit(i_win)
        if target_index < 0:
            self.add_tab()
        self.tabs[target_index].append(win, title)
        win.rect = old_rect
        if layout := win._parent_layout_ref():
            layout.remove(win)
        self.tabs.current_index = i_tab
        return None

    def _window_activated(self):
        back = self._backend_main_window
        back._update_context()
        i_tab = back._current_tab_index()
        if i_tab is None:
            return back._update_control_widget(None)
        tab = self.tabs.get(i_tab)
        if tab is None or len(tab) == 0:
            return back._update_control_widget(None)
        i_win = back._current_sub_window_index(i_tab)
        if i_win is None or len(tab) <= i_win:
            return back._update_control_widget(None)
        win = tab[i_win]
        back._update_control_widget(win.widget)
        _checker.call_widget_activated_callback(win.widget)
        self.events.window_activated.emit(win)

    def _main_window_resized(self, size: Size):
        if tab := self.tabs.current():
            for layout in tab.layouts:
                layout.anchor = layout.anchor.update_for_window_rect(size, layout.rect)
                layout._reanchor(size)
            for win in tab:
                win.anchor = win.anchor.update_for_window_rect(size, win.rect)
                win._reanchor(size)

    @contextmanager
    def _execute_in_context(
        self,
        is_gui: bool = False,
        process_model_output: bool = True,
        unwrap_future: bool = True,
    ):
        with self._global_lock:
            old_inst = self._instructions.model_copy()
            self._instructions = self._instructions.updated(
                gui_execution=is_gui,
                process_model_output=process_model_output,
                unwrap_future=unwrap_future,
            )
            try:
                yield None
            finally:
                self._instructions = old_inst

    def _iter_widget_class(self, model: WidgetDataModel) -> Iterator[type[_W]]:
        """Pick the most suitable widget class for the given model."""
        if model.force_open_with:
            yield import_object(model.force_open_with)
            return
        widget_classes, fallback_class = self._backend_main_window._list_widget_class(
            model.type
        )
        if not widget_classes:
            warnings.warn(
                f"No widget class is registered for model type {model.type!r}.",
                RuntimeWarning,
                stacklevel=2,
            )
            yield fallback_class
            return
        complete_match = [
            (tup.priority, tup.widget_class)
            for tup in widget_classes
            if tup.type == model.type and tup.priority >= 0
        ]
        if complete_match:
            yield from _iter_sorted(complete_match)
        subtype_match = [
            ((tup.type.count("."), tup.priority), tup.widget_class)
            for tup in widget_classes
            if tup.priority >= 0
        ]
        yield from _iter_sorted(subtype_match)

    def _pick_widget(self, model: WidgetDataModel) -> _W:
        """Pick the most suitable widget for the given model."""
        exceptions: list[tuple[Any, Exception]] = []
        for factory in self._iter_widget_class(model):
            try:
                try:
                    widget = factory(self)
                except TypeError:
                    widget = factory()
                widget_id = get_widget_class_id(type(widget))
                reg = _actions.AppActionRegistry.instance()
                if self.model_app.name != "." and (
                    plugin_configs := self.app_profile.plugin_configs.get(widget_id)
                ):
                    params = {}
                    for k, v in plugin_configs.items():
                        params[k] = v["value"]
                    cfgs = reg._plugin_default_configs
                    cfg_type = cfgs[widget_id].config_class
                    # widget should always have `update_configs` in this case
                    widget.update_configs(cfg_type(**params))
                widget.update_model(model)
            except Exception as e:
                exceptions.append((factory, e))
            else:
                break
        else:
            raise ValueError(
                f"Failed to create a widget for {_short_repr(model)}. Errors:\n"
                f"{_format_exceptions(exceptions)}"
            ) from exceptions[-1][1]
        if exceptions:
            raise exceptions[-1][1]

        return widget

    def _on_command_execution(self, id: str, result: Future):
        if exc := result.exception():
            _LOGGER.exception("Command %r failed: %r", id, exc)
            return
        if action := self.model_app.registered_actions.get(id):
            if getattr(action.callback, NO_RECORDING_FIELD, False):
                return None
            self._history_command.add(id)


def _short_repr(obj: Any) -> str:
    obj_repr = repr(obj)
    if len(obj_repr) > 50:
        obj_repr = obj_repr[:50] + "..."
    return obj_repr


def _format_exceptions(exceptions: list[tuple[Any, Exception]]) -> str:
    strs: list[str] = []
    for factory, e in exceptions:
        fname = getattr(factory, "__himena_widget_id__", repr(factory))
        strs.append(f"- {type(e).__name__} in {fname}\n  {e}")
    return "\n".join(strs)


def _iter_sorted(matched: list[tuple[tuple[int, int], type]]) -> Iterator[type]:
    _s = sorted(matched, key=lambda x: x[0], reverse=True)
    yield from (cls for _, cls in _s)



from __future__ import annotations
import sys
from contextlib import suppress
from typing import TYPE_CHECKING
import textwrap
import logging
from qtpy import QtWidgets as QtW, QtGui
from qtpy.QtCore import Qt, Signal
from himena.qt._qfinderwidget import QFinderWidget
from himena.consts import MonospaceFontFamily
from himena.plugins import validate_protocol

if TYPE_CHECKING:
    from himena_builtins.qt.output import OutputConfig


class QLogger(QtW.QPlainTextEdit):
    process = Signal(str)

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.setFont(QtGui.QFont(MonospaceFontFamily, 8))
        self.setReadOnly(True)
        self.setWordWrapMode(QtGui.QTextOption.WrapMode.NoWrap)
        self.process.connect(self.update_text)

        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)

        self._finder_widget = None

    def update_text(self, obj: str):
        text = "\n".join(textwrap.wrap(obj, width=200))
        if obj.endswith("\n"):
            text += "\n"
        self.moveCursor(QtGui.QTextCursor.MoveOperation.End)
        self.insertPlainText(obj)
        self.moveCursor(QtGui.QTextCursor.MoveOperation.End)
        return None

    def appendText(self, text: str):
        """Append text in the main thread."""
        self._emit_output(text)

    def _emit_output(self, text: str):
        with suppress(RuntimeError, OSError):
            self.process.emit(text)

    def _find_string(self):
        if self._finder_widget is None:
            self._finder_widget = QFinderWidget(self)
        self._finder_widget.show()
        self._align_finder()

    def resizeEvent(self, event):
        if self._finder_widget is not None:
            self._align_finder()
        super().resizeEvent(event)

    def _align_finder(self):
        if fd := self._finder_widget:
            vbar = self.verticalScrollBar()
            if vbar.isVisible():
                fd.move(self.width() - fd.width() - vbar.width() - 3, 5)
            else:
                fd.move(self.width() - fd.width() - 3, 5)

    def keyPressEvent(self, e: QtGui.QKeyEvent | None) -> None:
        if (
            e.key() == Qt.Key.Key_F
            and e.modifiers() & Qt.KeyboardModifier.ControlModifier
        ):
            self._find_string()
            return None
        return super().keyPressEvent(e)


class QtOutputWidget(QtW.QTabWidget):
    log_level_changed = Signal(str)
    log_filter_changed = Signal(str)

    def __init__(self):
        super().__init__()
        # stdout
        stdout_container = QtW.QWidget()
        self._stdout = QLogger()
        layout = QtW.QVBoxLayout(stdout_container)
        layout.addWidget(self._stdout)

        # logging
        logger_container = QtW.QWidget()
        self._log_level = QtW.QComboBox()
        self._log_level.addItems(["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"])
        self._log_level.setCurrentIndex(1)
        self._log_filter = QtW.QLineEdit()
        self._log_filter.setPlaceholderText("Log filter text")
        self._log_filter.editingFinished.connect(
            lambda: self.log_filter_changed.emit(self._log_filter.text())
        )
        self._logger = QLogger()
        layout = QtW.QVBoxLayout(logger_container)
        layout.addWidget(self._log_level)
        layout.addWidget(self._log_filter)
        layout.addWidget(self._logger)
        self._log_level.currentTextChanged.connect(self.log_level_changed.emit)

        # add tabs
        self.addTab(stdout_container, "stdout")
        self.addTab(logger_container, "log")


class _io_handler:
    def __init__(self, logger: QLogger):
        self._logger = logger

    def write(self, msg) -> None:
        """Handle the print event."""
        self._logger.appendText(msg)

    def flush(self):
        """Do nothing."""


class OutputInterface(logging.Handler):
    """A widget for displaying the standard output and logs.

    When this widget is opened, all the outputs to the standard output and loggings
    are redirected to this widget. In the logging tab, you can filter the logs by
    the log level and the module name.
    """

    _instances: dict[str, OutputInterface] = {}

    def __init__(self):
        super().__init__()
        self._widget = QtOutputWidget()
        self._logger = logging.getLogger()
        self._stdout_handler = _io_handler(self._widget._stdout)
        self._default_handlers = self._logger.handlers.copy()
        self._logger.setLevel(logging.INFO)
        self._widget.log_level_changed.connect(self.set_log_level)
        self._widget.log_filter_changed.connect(self.set_log_filter)

    def emit(self, record: logging.LogRecord):
        """Handle the logging event."""
        log_entry = self.format(record)
        self._widget._logger.appendText(f"{record.levelname}: {log_entry}\n")
        return None

    def set_log_level(self, level: str):
        log_level = getattr(logging, level)
        self._logger.setLevel(log_level)
        return None

    def set_log_filter(self, text: str):
        text = text.strip()
        for hnd in self._logger.handlers:
            hnd.filters.clear()
            if text:
                hnd.addFilter(logging.Filter(text))

    def connect_stdout(self):
        sys.stdout = self._stdout_handler

    def disconnect_stdout(self):
        sys.stdout = sys.__stdout__

    def close(self):
        self.disconnect_logger()

    def connect_logger(self):
        default = self._logger
        for handler in default.handlers:
            default.removeHandler(handler)
        default.addHandler(self)

    def disconnect_logger(self):
        default = self._logger
        for handler in default.handlers:
            default.removeHandler(handler)
        for handler in self._default_handlers:
            default.addHandler(handler)

    @validate_protocol
    def native_widget(self) -> QtW.QWidget:
        return self._widget

    def update_configs(self, cfg: OutputConfig):
        self.setFormatter(logging.Formatter(fmt=cfg.format, datefmt=cfg.date_format))


def get_widget(id: str = "default") -> OutputInterface:
    interf = OutputInterface()
    interf.connect_stdout()
    interf.connect_logger()
    OutputInterface._instances[id] = interf
    return interf



"""`himena` uses profile name as if it is a subcommand.

Examples
--------
$ himena  # launch GUI with the default profile
$ himena myprof  # launch GUI with the profile named "myprof"
$ himena path/to/file.txt  # open the file with the default profile
$ himena myprof path/to/file.txt  # open the file with the profile named "myprof"

"""

import logging
from pathlib import Path
from typing import TYPE_CHECKING
import sys
from himena._cli import HimenaArgumentParser, HimenaCliNamespace
from himena._socket import (
    SocketInfo,
    get_unique_lock_file,
    lock_file_path,
)

if TYPE_CHECKING:
    from himena.profile import AppProfile
    from himena.widgets import MainWindow
    from io import TextIOWrapper


def _is_testing() -> bool:
    return "pytest" in sys.modules


def _main(args: HimenaCliNamespace):
    if args.remove:
        from himena._cli.profiles import remove_profile

        return remove_profile(args)
    if args.new:
        from himena._cli.profiles import new_profile

        return new_profile(args)

    args.norm_profile_and_path()

    from himena.profile import load_app_profile

    prof_name = args.profile or "default"

    if args.clear_plugin_configs:
        prof = load_app_profile(prof_name)
        if prof.plugin_configs:
            prof.plugin_configs.clear()
            prof.save()
            print(f"Plugin configurations are cleared for the profile {prof_name!r}.")

    if args.uninstall_outdated:
        from himena._cli.install import uninstall_outdated

        return uninstall_outdated(prof_name)

    if args.list_plugins:
        from himena.utils.entries import iter_plugin_info

        app_profile = load_app_profile(prof_name)
        print("Profile:", prof_name)
        print("Plugins:")
        for info in iter_plugin_info():
            if info.place in app_profile.plugins:
                print(f"- {info.name} ({info.place}, v{info.version})")
        return

    if args.list_profiles:
        from himena.profile import profile_dir

        for path in profile_dir().iterdir():
            print(path.stem)
        return

    if args.get:
        from himena._cli.install import get_and_install

        return get_and_install(args.get, prof_name)

    if args.install or args.uninstall:
        from himena._cli.install import install_and_uninstall

        return install_and_uninstall(args.install, args.uninstall, prof_name)

    logging.basicConfig(level=args.log_level)

    # now it's ready to start the GUI
    app_prof = load_app_profile(prof_name, create_default=args.profile is None)
    attrs = {
        "print_import_time": args.import_time,
        "host": args.host,
        "port": args.port,
    }

    ui, lock = _send_or_create_window(app_prof, args.abs_path(), attrs)
    if ui is not None:
        ui.show(run=not _is_testing())
        lock.close()
        Path(lock.name).unlink(missing_ok=True)


def _send_or_create_window(
    prof: "AppProfile",
    path: str | None = None,
    attrs: dict = {},
) -> "tuple[MainWindow, TextIOWrapper] | None":
    from himena.core import new_window

    if (lock := get_unique_lock_file(prof.name)) is None:
        socket_info = SocketInfo.from_lock(prof.name)
        if path:
            files = path.split(";")
        else:
            files = []
        succeeded = socket_info.send_to_window(prof.name, files)
        if succeeded:
            return None, None
        lock = lock_file_path(prof.name).open("w")

    ui = new_window(prof, app_attributes=attrs)
    ui.socket_info.dump(lock)
    if path is not None:
        ui.read_file(path)
    return ui, lock


def main():
    parser = HimenaArgumentParser()

    # Run the main function with the parsed arguments
    args = parser.parse_args()
    _main(args)

    from himena.widgets._initialize import cleanup

    cleanup()



from typing import Any, Iterator, TYPE_CHECKING
from datetime import datetime as _datetime
import uuid

from pydantic import BaseModel, Field


if TYPE_CHECKING:
    from himena.types import WidgetDataModel
    from himena.workflow import Workflow
    import in_n_out as ino


class WorkflowStep(BaseModel):
    """The base class for a single step in a workflow."""

    type: str
    """Literal string that describes the type of the instance."""

    datetime: _datetime = Field(default_factory=_datetime.now)
    """The timestamp of the creation of this instance."""

    id: uuid.UUID = Field(default_factory=lambda: uuid.uuid4())
    """The unique identifier of the workflow step across runtime."""

    process_output: bool = Field(default=False)
    """Whether the output of this step should be processed by application."""

    def iter_parents(self) -> Iterator[uuid.UUID]:
        raise NotImplementedError("This method must be implemented in a subclass.")

    def _get_model_impl(self, wf: "Workflow") -> "WidgetDataModel":
        raise NotImplementedError("This method must be implemented in a subclass.")

    def get_model(
        self,
        wf: "Workflow",
        *,
        force_process_output: bool = False,
        metadata: Any | None = None,
    ) -> "WidgetDataModel":
        if win := wf._mock_main_window.window_for_id(self.id):
            out = win.to_model()
            return out
        model = self._get_model_impl(wf)
        model.workflow = wf
        if metadata is not None:
            model.metadata = metadata
        if self.process_output or force_process_output:
            self._current_store().process(model)
        return model

    def __repr_args__(self):  # simplify the repr output
        for arg in super().__repr_args__():
            if arg[0] in ("type", "datetime", "id"):
                continue
            yield arg

    def __str__(self):
        return repr(self)

    def _current_store(self) -> "ino.Store":
        from himena.widgets import current_instance

        return current_instance().model_app.injection_store

    def construct_workflow(self) -> "Workflow":
        from himena.workflow import Workflow

        return Workflow(steps=[self])



from pathlib import Path
from typing import Any, TYPE_CHECKING
import uuid
from himena.workflow._graph import Workflow
from himena.workflow._reader import LocalReaderMethod, UserInput, RuntimeInputBound

if TYPE_CHECKING:
    from himena.widgets import MainWindow


def workflow_to_input_fields(wf: Workflow) -> dict[uuid.UUID, Any]:
    from himena.types import WidgetDataModel

    fields = {}
    for step in wf.steps:
        if not isinstance(step, UserInput):
            continue
        option = {"label": step.label, "tooltip": step.doc}
        fields[step.id] = option
        if step.how == "file":
            option["annotation"] = Path
        elif step.how == "model":
            option["annotation"] = WidgetDataModel
        else:
            raise NotImplementedError(f"Unsupported input type: {step.how}")

    return fields


def as_function(wf: Workflow):
    """Convert a workflow to a function that is ready to be registered as an action."""
    from himena.types import Parametric, WidgetDataModel
    from himena.plugins import configure_gui

    fields = workflow_to_input_fields(wf)
    options = {f"arg{i}": val for i, val in enumerate(fields.values())}
    key_to_id = {f"arg{i}": k for i, k in enumerate(fields.keys())}

    def func(ui: "MainWindow") -> Parametric:
        @configure_gui(**options)
        def inner(**kwargs):
            wf_compute = wf.model_copy()
            wf_out = wf_compute.model_copy()
            for k, v in kwargs.items():
                step_id = key_to_id[k]
                if isinstance(v, Path):
                    meth = LocalReaderMethod(id=step_id, path=v)
                    wf_compute = wf_compute.replace(step_id, meth)
                    wf_out = wf_out.replace(step_id, meth)
                elif isinstance(v, WidgetDataModel):
                    wf_compute = wf_compute.replace(
                        step_id, RuntimeInputBound(id=step_id, bound_value=v)
                    )
                    wf_out = wf_out.replace(step_id, v.workflow)
                else:
                    raise NotImplementedError(f"Unsupported type: {type(v)}")
            out = wf_compute.compute(process_output=True)
            if isinstance(out, WidgetDataModel):
                ui.current_window._update_model_workflow(wf_out)
            return None

        return inner

    return func


def as_function_from_path(path):
    """Convert a workflow file to a function ready to be registered as an action."""
    wf = Workflow.model_validate_json(Path(path).read_text())
    return as_function(wf)



import pytest
from himena.qt.magicgui._basic_widgets import float_to_str

@pytest.mark.parametrize(
    "value, expected",
    [
        (1.0, "1.0"),
        (14, "14"),
        (0.024, "0.024"),
        (0.046999999998, "0.047"),
        (1.32e8, "1.32e+08"),
        (0.00005335, "5.335e-05"),
        (199000000, "1.99e+08"),
    ]
)
def test_float_to_str(value, expected: str):
    assert float_to_str(value) == expected



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Iterator, NamedTuple
from psygnal import Signal


if TYPE_CHECKING:
    Range = tuple[slice[int, int, None], slice[int, int, None]]


class Index(NamedTuple):
    """Index tuple,"""

    row: int
    column: int

    def as_uint(self) -> Index:
        """Return unsigned Index"""
        return Index(max(self.row, 0), max(self.column, 0))


class DummyRange(NamedTuple):
    row: slice
    column: slice


_DUMMY_RANGE = DummyRange(slice(0, 0), slice(0, 0))


class SelectionModel:
    """A specialized range model with item-selection-like behavior."""

    moving = Signal(Index, Index)
    moved = Signal(Index, Index)

    def __init__(self, row_count: Callable[[], int], col_count: Callable[[], int]):
        self._ranges: list[Range] = []
        self._is_blocked = False
        self._selected_indices: set[int] = set()

        # indices of self._ranges that are row/column selections
        self._row_selection_indices: set[int] = set()
        self._col_selection_indices: set[int] = set()

        self._ctrl_on = False
        self._shift_on = False
        self._selection_start: Index | None = None
        self._current_index = Index(0, 0)
        self._row_count_getter = row_count
        self._col_count_getter = col_count

    @property
    def current_index(self) -> Index:
        """Current position of the selection cursor."""
        return self._current_index

    @current_index.setter
    def current_index(self, index: tuple[int, int]):
        self._current_index = Index(*index)

    @property
    def current_range(self) -> tuple[slice, slice] | None:
        if len(self._ranges) > 0:
            return self._ranges[-1]
        return None

    @property
    def start(self) -> Index | None:
        """The selection starting index."""
        return self._selection_start

    def __len__(self) -> int:
        """Number of ranges"""
        return len(self._ranges)

    def __iter__(self) -> Iterator[Range]:
        """Iterate over all the ranges."""
        return iter(self._ranges)

    def __getitem__(self, index: int) -> Range:
        """Get the range at the specified index."""
        return self._ranges[index]

    @property
    def ranges(self) -> list[Range]:
        return list(self._ranges)

    def get_single_range(self) -> Range:
        """Return the only range in the selection model."""
        if len(self._ranges) == 0:
            r, c = self.current_index
            return (slice(r, r + 1), slice(c, c + 1))
        elif len(self._ranges) == 1:
            return self._ranges[0]
        raise ValueError("Multiple ranges are selected.")

    def iter_row_selections(self) -> Iterator[slice]:
        for i in self._row_selection_indices:
            yield self._ranges[i][0]

    def iter_col_selections(self) -> Iterator[slice]:
        for i in self._col_selection_indices:
            yield self._ranges[i][1]

    def num_row_selections(self) -> int:
        return len(self._row_selection_indices)

    def num_col_selections(self) -> int:
        return len(self._col_selection_indices)

    def append(self, range: Range, row: bool = False, column: bool = False) -> None:
        """Append a new range."""
        if self._is_blocked:
            return None
        self._ranges.append(range)
        if row:
            self._row_selection_indices.add(len(self._ranges) - 1)
        elif column:
            self._col_selection_indices.add(len(self._ranges) - 1)
        return None

    def update_last(self, range: Range, row: bool = False, col: bool = False) -> None:
        """Update the last range with new one."""
        if self._is_blocked:
            return None
        if self._ranges:
            self._ranges[-1] = range
        else:
            self._ranges.append(range)
        if row:
            self._row_selection_indices.add(len(self._ranges) - 1)
        elif col:
            self._col_selection_indices.add(len(self._ranges) - 1)
        return None

    def contains(self, index: tuple[int, int]) -> bool:
        """Whether the index is in the selection."""
        r0, c0 = index
        for r, c in self._ranges:
            if r.start <= r0 < r.stop and c.start <= c0 < c.stop:
                return True
        return False

    def set_ranges(self, ranges: list[Range]) -> None:
        if self._is_blocked:
            return None
        self.clear()
        return self._ranges.extend(ranges)

    def clear(self) -> None:
        """Clear all the selections"""
        if self._is_blocked:
            return None
        self._ranges.clear()
        self._row_selection_indices.clear()
        self._col_selection_indices.clear()
        return None

    def is_jumping(self) -> bool:
        """Whether the selection is jumping or not."""
        return len(self._ranges) > 0 and self._ranges[-1] is _DUMMY_RANGE

    def is_moving_to_edge(self) -> bool:
        """Whether the selection is moving to the edge by Ctrl+arrow key."""
        return not self.is_jumping() and self._ctrl_on

    def set_ctrl(self, on: bool) -> None:
        """Equivalent to pressing Ctrl."""
        self._ctrl_on = bool(on)
        return None

    def set_shift(self, on: bool) -> None:
        """Equivalent to pressing Shift."""
        self._shift_on = bool(on)
        if on and self._selection_start is None:
            self._selection_start = self._current_index
        return None

    def jump_to(self, r: int, c: int):
        """Emulate mouse click at cell (r, c)."""
        if self._ctrl_on and not self._shift_on:
            self._ranges.append(_DUMMY_RANGE)
        return self.move_to(r, c)

    def move_to(self, r: int, c: int):
        """Emulate dragging to cell (r, c)."""
        src = self._current_index
        dst = Index(r, c)
        self.moving.emit(src, dst)
        self._current_index = dst
        if self._is_blocked:
            return None

        if not self._shift_on:
            self._selection_start = None

        if self._selection_start is None:
            _r0 = _r1 = r
            _c0 = _c1 = c
        else:
            r0, c0 = self._selection_start
            _r0, _r1 = sorted([r0, r])
            _c0, _c1 = sorted([c0, c])

        if _r0 < 0:
            rsl = slice(0, self._row_count_getter())
            col = True
        else:
            rsl = slice(_r0, _r1 + 1)
            col = False
        if _c0 < 0:
            csl = slice(0, self._col_count_getter())
            row = True
        else:
            csl = slice(_c0, _c1 + 1)
            row = False

        if not self._shift_on:
            if not self.is_jumping():
                self.clear()
        elif self._selection_start is None:
            self._selection_start = self._current_index

        self.update_last((rsl, csl), row=row, col=col)
        self.moved.emit(src, dst)
        return None

    def move(self, dr: int, dc: int, allow_header: bool = False):
        """Move by (dr, dc) cells."""
        r, c = self._current_index
        idx_min = -int(allow_header)

        if dr != 0:
            nr = self._row_count_getter()
            r = min(r + dr, nr - 1)
            r = max(idx_min, r)

        if dc != 0:
            nc = self._col_count_getter()
            c = min(c + dc, nc - 1)
            c = max(idx_min, c)

        return self.move_to(r, c)

    def move_limited(self, dr: int, dc: int, nr: int, nc: int):
        """Move by (dr, dc) cells within the range of (nr, nc)."""
        r, c = self._current_index
        r = min(max(0, r + dr), nr - 1)
        c = min(max(0, c + dc), nc - 1)
        return self.move_to(r, c)



from himena import StandardType
from himena.plugins import register_widget_class
from himena_builtins.qt.widgets.ipynb import QIpynbEdit

register_widget_class(StandardType.IPYNB, QIpynbEdit, priority=50)



from pathlib import Path
from qtpy import QtGui
from qtpy.QtCore import Qt
from himena import MainWindow, StandardType
from himena.standards.model_meta import TextMeta
from himena.types import WidgetDataModel
from himena_builtins.qt.text import QTextEdit, QRichTextEdit, QSvgPreview, QMarkdownPreview
from pytestqt.qtbot import QtBot
from himena.testing import WidgetTester

_Ctrl = Qt.KeyboardModifier.ControlModifier


def test_text_edit(qtbot: QtBot):
    model = WidgetDataModel(value="a\nb", type="text")
    with WidgetTester(QTextEdit()) as tester:
        tester.update_model(model)
        qtbot.addWidget(tester.widget)
        main = tester.widget._main_text_edit
        tester.widget._control._wordwrap.setChecked(True)
        tester.widget._control._wordwrap.setChecked(False)

        assert tester.to_model().value == "a\nb"
        assert main.toPlainText() == "a\nb"
        # move to the end
        cursor = main.textCursor()
        cursor.setPosition(len(main.toPlainText()))
        main.setTextCursor(cursor)

        qtbot.keyClick(main, Qt.Key.Key_Return)
        qtbot.keyClick(main, Qt.Key.Key_Tab)
        qtbot.keyClick(main, Qt.Key.Key_Backtab)
        qtbot.keyClick(main, Qt.Key.Key_Tab)
        qtbot.keyClick(main, Qt.Key.Key_O)
        qtbot.keyClick(main, Qt.Key.Key_P)
        assert tester.to_model().value.splitlines() == ["a", "b", "    op"]
        qtbot.keyClick(main, Qt.Key.Key_Home)
        qtbot.keyClick(main, Qt.Key.Key_Up, modifier=Qt.KeyboardModifier.AltModifier)
        qtbot.keyClick(main, Qt.Key.Key_Down, modifier=Qt.KeyboardModifier.AltModifier)
        qtbot.keyClick(main, Qt.Key.Key_Down)
        qtbot.keyClick(main, Qt.Key.Key_Down)
        qtbot.keyClick(main, Qt.Key.Key_Down)
        qtbot.keyClick(main, Qt.Key.Key_Tab)
        qtbot.keyClick(main, Qt.Key.Key_X)
        qtbot.keyClick(main, Qt.Key.Key_Return)
        qtbot.keyClick(main, Qt.Key.Key_A)
        qtbot.keyClick(main, Qt.Key.Key_B)
        qtbot.keyClick(main, Qt.Key.Key_C)
        qtbot.keyClick(main, Qt.Key.Key_D)
        qtbot.keyClick(main, Qt.Key.Key_L, modifier=_Ctrl)
        qtbot.keyClick(main, Qt.Key.Key_Up, modifier=Qt.KeyboardModifier.AltModifier)
        qtbot.keyClick(main, Qt.Key.Key_Down, modifier=Qt.KeyboardModifier.AltModifier)
        qtbot.keyClick(main, Qt.Key.Key_Left)
        qtbot.keyClick(main, Qt.Key.Key_D, modifier=_Ctrl)
        qtbot.keyClick(main, Qt.Key.Key_C, modifier=_Ctrl)
        qtbot.keyClick(main, Qt.Key.Key_Return)
        qtbot.keyClick(main, Qt.Key.Key_V, modifier=_Ctrl)
        qtbot.keyClick(main, Qt.Key.Key_Less, modifier=_Ctrl)
        qtbot.keyClick(main, Qt.Key.Key_Greater, modifier=_Ctrl)
        qtbot.keyClick(main, Qt.Key.Key_Greater, modifier=_Ctrl)
        qtbot.keyClick(main, Qt.Key.Key_0, modifier=_Ctrl)
        qtbot.keyClick(tester.widget, Qt.Key.Key_F, modifier=_Ctrl)
        tester.widget.resize(100, 100)
        tester.widget.resize(120, 120)


def test_find_text(qtbot: QtBot):
    model = WidgetDataModel(value="a\nb\nc\nbc", type="text")
    text_edit = QTextEdit()
    text_edit.update_model(model)
    qtbot.addWidget(text_edit)
    qtbot.keyClick(text_edit, Qt.Key.Key_F, modifier=_Ctrl)
    finder = text_edit._main_text_edit._finder_widget
    assert finder is not None
    finder._line_edit.setText("b")
    qtbot.keyClick(finder, Qt.Key.Key_Enter)
    qtbot.keyClick(finder, Qt.Key.Key_Enter, modifier=Qt.KeyboardModifier.ShiftModifier)
    finder._btn_next.click()
    finder._btn_prev.click()

def test_svg_preview(sample_dir: Path, qtbot: QtBot):
    with WidgetTester(QSvgPreview()) as tester:
        qtbot.addWidget(tester.widget)
        svg_path = sample_dir / "svg.svg"
        tester.update_model(value=svg_path.read_text(), type=StandardType.SVG)
        tester.to_model()

def test_markdow_preview(sample_dir: Path, qtbot: QtBot):
    with WidgetTester(QMarkdownPreview()) as tester:
        qtbot.addWidget(tester.widget)
        md_path = sample_dir / "markdown.md"
        tester.update_model(value=md_path.read_text(), type=StandardType.MARKDOWN)
        tester.to_model()

def test_rich_text(sample_dir: Path, qtbot: QtBot):
    with WidgetTester(QRichTextEdit()) as tester:
        qtbot.addWidget(tester.widget)
        md_path = sample_dir / "html.html"
        tester.update_model(value=md_path.read_text(), type=StandardType.HTML)
        tester.to_model()
        tester.widget._control._on_foreground_color_changed(QtGui.QColor("blue"))
        tester.widget._control._on_background_color_changed(QtGui.QColor("red"))
        tester.widget._control._on_toggle_bold()
        tester.widget._control._on_toggle_italic()
        tester.widget._control._on_toggle_underline()
        tester.widget._control._on_toggle_strike()

def test_commands(himena_ui: MainWindow):
    win = himena_ui.add_object("print(2)", type=StandardType.TEXT)
    win.update_model(win.to_model().with_metadata(TextMeta(language="python")))
    himena_ui.exec_action("builtins:text:run-script")
    win = himena_ui.add_object("1,2,3", type=StandardType.TEXT)
    himena_ui.exec_action("builtins:text:change-separator", with_params={})
    assert himena_ui.current_model.value == "1\t2\t3"
    himena_ui.exec_action("builtins:text:change-encoding", with_params={"encoding": "utf-16"})
    assert himena_ui.current_model.metadata.encoding == "utf-16"
    himena_ui.add_object("def f(x):\n    return x + 1", type=StandardType.TEXT)
    himena_ui.exec_action("builtins:text:compile-as-function")
    himena_ui.add_object("def f(x):\n    return x + 1\nf", type=StandardType.TEXT)
    himena_ui.exec_action("builtins:text:compile-as-function")
    himena_ui.add_object("def main(x):\n    return x\n", type=StandardType.TEXT)
    himena_ui.exec_action("builtins:text:run-script-main", with_params={"x": himena_ui.tabs[0][0]})

def test_open_as_text_anyway(sample_dir: Path, himena_ui: MainWindow):
    himena_ui.read_file(sample_dir / "random_ext.aaa")
    assert himena_ui.current_model.type == StandardType.READER_NOT_FOUND
    himena_ui.exec_action("builtins:open-as-text-anyway")
    assert len(himena_ui.tabs[0]) == 1
    assert himena_ui.current_model.type == StandardType.TEXT
    assert himena_ui.current_model.value == sample_dir.joinpath("random_ext.aaa").read_text()
    assert himena_ui.current_model.extension_default == ".aaa"



import pytest
from pathlib import Path
from himena.consts import StandardType
from himena.io_utils import read, write
from himena.standards import roi
from himena.types import WidgetDataModel
from himena_builtins.tools.others import show_statistics, show_metadata
from himena_builtins.io import read_as_pandas_dataframe, read_as_polars_dataframe, read_as_text_anyway

@pytest.mark.parametrize(
    "file_name, model_type",
    [
        ("text.txt", StandardType.TEXT),
        ("json.json", StandardType.JSON),
        ("svg.svg", StandardType.SVG),
        ("table.csv", StandardType.TABLE),
        ("table_nonuniform.csv", StandardType.TABLE),
        ("image.png", StandardType.IMAGE),
        ("html.html", StandardType.HTML),
        ("excel.xlsx", StandardType.EXCEL),
        ("array.npy", StandardType.ARRAY),
        ("array_structured.npy", StandardType.ARRAY),
        ("ipynb.ipynb", StandardType.IPYNB),
    ]
)
def test_reading_writing_files(sample_dir: Path, tmpdir, file_name: str, model_type: str):
    tmpdir = Path(tmpdir)
    model = read(sample_dir / file_name)
    assert model.type == model_type
    write(model, tmpdir / file_name)
    model = read(sample_dir / file_name)
    assert model.type == model_type
    show_statistics(model)
    show_metadata(model)

def test_dataframe(sample_dir: Path, tmpdir):
    tmpdir = Path(tmpdir)
    model = read(sample_dir / "pq.parquet")
    assert model.type == StandardType.DATAFRAME
    write(model, tmpdir / "pq.parquet")
    model = read(sample_dir / "pq.parquet")
    assert model.type == StandardType.DATAFRAME
    show_statistics(model)
    show_metadata(model)

def test_pandas_io(tmpdir):
    import pandas as pd

    df = pd.DataFrame({"A": [1, 2, 3], "B": ["a", "b", "c"]})
    save_dir = Path(tmpdir)
    df.to_csv(save_dir / "test.csv", index=False)
    read_as_pandas_dataframe(save_dir / "test.csv")
    df.to_csv(save_dir / "test.tsv", index=False, sep="\t")
    read_as_pandas_dataframe(save_dir / "test.tsv")
    df.to_json(save_dir / "test.json", index=False)
    read_as_pandas_dataframe(save_dir / "test.json")
    df.to_parquet(save_dir / "test.parquet", index=False)
    read_as_pandas_dataframe(save_dir / "test.parquet")
    df.to_feather(save_dir / "test.feather")
    read_as_pandas_dataframe(save_dir / "test.feather")

def test_polars_io(tmpdir):
    import polars as pl

    df = pl.DataFrame({"A": [1, 2, 3], "B": ["a", "b", "c"]})
    save_dir = Path(tmpdir)
    df.write_csv(save_dir / "test.csv")
    read_as_polars_dataframe(save_dir / "test.csv")
    df.write_csv(save_dir / "test.tsv", separator="\t")
    read_as_polars_dataframe(save_dir / "test.tsv")
    df.write_parquet(save_dir / "test.parquet")
    read_as_polars_dataframe(save_dir / "test.parquet")
    df.write_json(save_dir / "test.json")
    read_as_polars_dataframe(save_dir / "test.json")
    df.write_ipc(save_dir / "test.feather")
    read_as_polars_dataframe(save_dir / "test.feather")

def test_read_as_text_anyway(tmpdir):
    text = "This is a test."
    file_path = Path(tmpdir, "test.xxx")
    file_path.write_text(text)
    result = read_as_text_anyway(file_path)
    assert result.value == text

def test_roi_reader(tmpdir):
    file_path = Path(tmpdir, "roi.roi")
    roi_list = roi.RoiListModel(
        items=[
            roi.RectangleRoi(x=1, y=2, width=3, height=4),
            roi.EllipseRoi(x=5, y=6, width=7, height=8),
        ]
    )
    write(WidgetDataModel(value=roi_list, type=StandardType.ROIS), file_path)
    read(file_path)



from __future__ import annotations

from qtpy import QtWidgets as QtW, QtCore
from typing import TYPE_CHECKING
from himena.qt.settings._shared import QInstruction

if TYPE_CHECKING:
    from himena.widgets import MainWindow


class QStartupCommandsPanel(QtW.QWidget):
    """Widget to edit the startup commands."""

    def __init__(self, ui: MainWindow):
        super().__init__()
        self._ui = ui
        self._instruction = QInstruction(
            "List command IDs to be executed at startup.",
        )
        self._text_edit = QtW.QPlainTextEdit(self)
        self._text_edit.setPlainText("\n".join(self._ui.app_profile.startup_commands))
        self._footer = QtW.QWidget(self)
        _footer_layout = QtW.QHBoxLayout(self._footer)
        _footer_layout.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)
        self._apply_button = QtW.QPushButton("Apply", self._footer)
        self._apply_button.clicked.connect(self._apply_changes)
        self._msg_label = QtW.QLabel(self._footer)
        self._msg_label.setMaximumWidth(280)
        _footer_layout.setContentsMargins(0, 0, 0, 0)
        _footer_layout.addWidget(self._msg_label)
        _footer_layout.addWidget(self._apply_button)
        layout = QtW.QVBoxLayout(self)
        layout.addWidget(self._instruction)
        layout.addWidget(self._text_edit)
        layout.addWidget(self._footer)
        self._text_edit.textChanged.connect(self._on_text_changed)

    def _apply_changes(self):
        commands: list[str] = []
        commands_unknown: list[str] = []
        app = self._ui.model_app
        for line in self._text_edit.toPlainText().splitlines():
            line = line.strip()
            if line not in app.commands:
                commands_unknown.append(line)
            commands.append(line)
        if commands_unknown:
            self._msg_label.setText(f"Unknown commands: {', '.join(commands_unknown)}")
        else:
            self._msg_label.setText("Changes applied.")
        self._ui.app_profile.startup_commands = commands
        self._ui.app_profile.save()

    def _on_text_changed(self):
        self._msg_label.clear()



from __future__ import annotations

from contextlib import suppress
import logging
import math
from typing import Iterable
import numpy as np
from qtpy import QtWidgets as QtW, QtCore, QtGui
from qtpy.QtCore import Qt

from himena.consts import DefaultFontFamily
from himena.types import Size

from ._base import QBaseGraphicsView, QBaseGraphicsScene
from ._roi_items import (
    QRoi,
    QRectangleRoi,
    ROI_MODES,
    MouseMode,
)
from ._handles import QHandleRect, RoiSelectionHandles
from ._scale_bar import QScaleBarItem
from himena_builtins.qt.widgets._image_components import _mouse_events as _me
from himena.qt import ndarray_to_qimage
from himena.widgets import show_tooltip, get_clipboard, set_clipboard, current_instance

_LOGGER = logging.getLogger(__name__)


class QImageGraphicsWidget(QtW.QGraphicsWidget):
    def __init__(self, parent=None, additive: bool = False):
        super().__init__(parent)
        self._img: np.ndarray = np.zeros((0, 0))
        self._qimage = QtGui.QImage()
        self._smoothing = False
        self.set_additive(additive)

    def set_image(self, img: np.ndarray):
        """Set a (colored) image to display."""
        qimg = ndarray_to_qimage(img)
        self._img = img
        self._qimage = qimg
        self.update()

    def set_additive(self, additive: bool):
        if additive:
            self._comp_mode = QtGui.QPainter.CompositionMode.CompositionMode_Plus
        else:
            self._comp_mode = QtGui.QPainter.CompositionMode.CompositionMode_SourceOver

    def setSmoothingEnabled(self, enabled):
        self._smoothing = enabled
        self.update()

    def initPainter(self, painter: QtGui.QPainter):
        painter.setCompositionMode(self._comp_mode)
        painter.setRenderHint(
            QtGui.QPainter.RenderHint.SmoothPixmapTransform, self._smoothing
        )

    def paint(self, painter, option, widget=None):
        if self._qimage.isNull():
            return

        self.initPainter(painter)
        bounding_rect = self.boundingRect()
        painter.drawImage(bounding_rect, self._qimage)
        is_light_bg = (
            self.scene().views()[0].backgroundBrush().color().lightness() > 128
        )
        if is_light_bg:
            pen = QtGui.QPen(QtGui.QColor(19, 19, 19), 1)
        else:
            pen = QtGui.QPen(QtGui.QColor(236, 236, 236), 1)
        pen.setCosmetic(True)
        painter.setPen(pen)
        painter.drawRect(bounding_rect)

    def boundingRect(self):
        height, width = self._img.shape[:2]
        return QtCore.QRectF(0, 0, width, height)


class QRoiLabels(QtW.QGraphicsItem):
    """Item that shows labels for ROIs in the paint method"""

    def __init__(self, view: QImageGraphicsView):
        super().__init__()
        self._view = view
        self._show_labels = False
        self._font = QtGui.QFont(DefaultFontFamily, 10)
        self._bounding_rect = QtCore.QRectF(0, 0, 0, 0)

    def paint(
        self,
        painter: QtGui.QPainter,
        option: QtW.QStyleOptionGraphicsItem,
        widget: QtW.QWidget,
    ):
        if not self._show_labels:
            return
        if not self._view._is_rois_visible:
            return
        scale = self.scene().views()[0].transform().m11()
        self._font.setPointSizeF(9 / scale)
        painter.setFont(self._font)
        metrics = QtGui.QFontMetricsF(self._font)
        for ith, roi in enumerate(self._view._roi_items):
            pos = self.mapToScene(roi.boundingRect().center())
            roi_label = roi.label() or str(ith)
            width = metrics.width(roi_label)
            height = metrics.height()
            painter.setPen(QtGui.QPen(QtGui.QColor(0, 0, 0, 0), 1))
            painter.setBrush(QtGui.QBrush(QtGui.QColor(0, 0, 0, 128)))
            rect = QtCore.QRectF(
                pos.x() - width / 2, pos.y() - height / 2, width, height
            )
            painter.drawRect(rect.adjusted(-0.3, 0, 0.3, 0))
            painter.setPen(QtGui.QPen(QtGui.QColor(255, 255, 255), 1))
            painter.setBrush(QtGui.QBrush(QtGui.QColor(255, 255, 255)))
            painter.drawText(rect, roi_label)

    def boundingRect(self):
        return self._bounding_rect

    def set_bounding_rect(self, rect: QtCore.QRectF):
        self._bounding_rect = QtCore.QRectF(rect)
        self.update()


class QImageGraphicsView(QBaseGraphicsView):
    roi_added = QtCore.Signal(QRoi)
    roi_removed = QtCore.Signal(int)
    """Emitted when a i-th ROI *in this slice* is removed from the view."""
    roi_visibility_changed = QtCore.Signal(bool)
    current_roi_updated = QtCore.Signal()
    """Emitted when the current ROI is added, removed or edited."""
    mode_changed = QtCore.Signal(MouseMode)
    hovered = QtCore.Signal(QtCore.QPointF)
    geometry_changed = QtCore.Signal(QtCore.QRectF)
    array_updated = QtCore.Signal(int, object)

    Mode = MouseMode

    def __init__(self, roi_visible: bool = False, roi_pen: QtGui.QPen | None = None):
        super().__init__()
        ### Attributes ###
        self._roi_items: list[QRoi] = []
        self._current_roi_item: QRoi | None = None
        self._is_current_roi_item_not_registered = False
        self._roi_pen = roi_pen or QtGui.QPen(QtGui.QColor(225, 225, 0), 3)
        self._roi_pen.setCosmetic(True)
        self._mode = MouseMode.PAN_ZOOM
        self._last_mode_before_key_hold = MouseMode.PAN_ZOOM
        self._mouse_event_handler = _me.PanZoomMouseEvents(self)
        self._is_drawing_multipoints = False
        self._is_rois_visible = roi_visible
        self._selection_handles = RoiSelectionHandles(self)
        self._initialized = False
        self._image_widgets: list[QImageGraphicsWidget] = []
        self.switch_mode(MouseMode.PAN_ZOOM)
        self._qroi_labels = self.addItem(QRoiLabels(self))
        self._qroi_labels.setZValue(10000)
        self._scale_bar_widget = self.addItem(QScaleBarItem(self))
        self._scale_bar_widget.setZValue(10000)
        self._scale_bar_widget.setVisible(False)
        self.geometry_changed.connect(self._scale_bar_widget.update_rect)
        self._stick_to_grid = False
        self.array_updated.connect(self._on_array_updated)
        self.setAttribute(QtCore.Qt.WidgetAttribute.WA_AcceptTouchEvents, True)

    def add_image_layer(self, additive: bool = False):
        self._image_widgets.append(
            self.addItem(QImageGraphicsWidget(additive=additive))
        )
        brect = self._image_widgets[0].boundingRect()
        self.scene().setSceneRect(brect)

    def set_n_images(self, num: int):
        if num < 1:
            raise ValueError("Number of images must be at least 1.")
        for _ in range(num - len(self._image_widgets)):
            additive = len(self._image_widgets) > 0
            self.add_image_layer(additive=additive)
        for _ in range(len(self._image_widgets) - num):
            widget = self._image_widgets.pop()
            self.scene().removeItem(widget)
            widget.deleteLater()

    def set_show_rois(self, show: bool):
        self._is_rois_visible = show
        for item in self._roi_items:
            item.setVisible(show)
        self._qroi_labels.update()
        self.roi_visibility_changed.emit(show)

    def set_show_labels(self, show: bool):
        self._qroi_labels._show_labels = show
        self._qroi_labels.update()

    def set_array(self, idx: int, img: np.ndarray | None):
        """Set an image to display."""
        # NOTE: image must be ready for conversion to QImage (uint8, mono or RGB)
        self.array_updated.emit(idx, img)

    def select_image(self):
        """Add a selection ROI to the entire image."""
        ny, nx = self._image_widgets[0]._img.shape[:2]
        self.remove_current_item(reason="Ctrl+A")
        self.set_current_roi(QRectangleRoi(0, 0, nx, ny).withPen(self._roi_pen))
        self._selection_handles.connect_rect(self._current_roi_item)
        return None

    def _on_array_updated(self, idx: int, img: np.ndarray | None):
        if idx >= len(self._image_widgets):
            return  # this happens when the number of channels decreased
        widget = self._image_widgets[idx]
        if img is None:
            widget.setVisible(False)
        else:
            widget.set_image(img)
            widget.setVisible(True)
        for widget in self._image_widgets:
            if widget.isVisible():
                self._qroi_labels.set_bounding_rect(widget.boundingRect())
                self._scale_bar_widget.set_bounding_rect(widget.boundingRect())
                break

    def set_image_blending(self, opaque: list[bool]):
        is_first = True
        for img, is_opaque in zip(self._image_widgets, opaque):
            if is_opaque and is_first:
                is_first = False
                img.set_additive(False)
            else:
                img.set_additive(True)

    def set_stick_to_grid(self, stick: bool):
        self._stick_to_grid = stick

    def clear_rois(self):
        scene = self.scene()
        for item in self._roi_items:
            scene.removeItem(item)
        self._roi_items.clear()
        if not self._is_current_roi_item_not_registered:
            self.remove_current_item(reason="clear all ROIs")

    def remove_rois(self, rois: Iterable[QRoi]):
        """Remove Qt ROIs from the view."""
        for roi in rois:
            if roi in self._roi_items:
                self._roi_items.remove(roi)
                self.scene().removeItem(roi)

    def extend_qrois(self, rois: Iterable[QRoi], current_roi: QRoi | None = None):
        """Set Qt ROIs to display."""
        for roi in rois:
            self.scene().addItem(roi)
            roi.setVisible(self._is_rois_visible)
            self._roi_items.append(roi)
            if roi is current_roi:
                self.select_item(
                    roi, is_registered_roi=not self._is_current_roi_item_not_registered
                )

    def mode(self) -> MouseMode:
        return self._mode

    def set_mode(self, mode: MouseMode):
        self._mode = mode
        _LOGGER.info("Mode changed to %r", mode)
        if mode in ROI_MODES:
            self.viewport().setCursor(Qt.CursorShape.CrossCursor)
            if mode is MouseMode.ROI_POINT:
                self._mouse_event_handler = _me.PointRoiMouseEvents(self)
            elif mode is MouseMode.ROI_POINTS:
                self._mouse_event_handler = _me.PointsRoiMouseEvents(self)
            elif mode is MouseMode.ROI_LINE:
                self._mouse_event_handler = _me.LineRoiMouseEvents(self)
            elif mode is MouseMode.ROI_RECTANGLE:
                self._mouse_event_handler = _me.RectangleRoiMouseEvents(self)
            elif mode is MouseMode.ROI_ROTATED_RECTANGLE:
                self._mouse_event_handler = _me.RotatedRectangleRoiMouseEvents(self)
            elif mode is MouseMode.ROI_ELLIPSE:
                self._mouse_event_handler = _me.EllipseRoiMouseEvents(self)
            elif mode is MouseMode.ROI_ROTATED_ELLIPSE:
                self._mouse_event_handler = _me.RotatedEllipseRoiMouseEvents(self)
            elif mode is MouseMode.ROI_CIRCLE:
                self._mouse_event_handler = _me.CircleRoiMouseEvents(self)
            elif mode is MouseMode.ROI_POLYGON:
                self._mouse_event_handler = _me.PolygonRoiMouseEvents(self)
            elif mode is MouseMode.ROI_SEGMENTED_LINE:
                self._mouse_event_handler = _me.SegmentedLineRoiMouseEvents(self)

        elif mode is MouseMode.SELECT:
            self.viewport().setCursor(Qt.CursorShape.ArrowCursor)
            self._mouse_event_handler = _me.SelectMouseEvents(self)
        elif mode is MouseMode.PAN_ZOOM:
            self.viewport().setCursor(Qt.CursorShape.ArrowCursor)
            self._mouse_event_handler = _me.PanZoomMouseEvents(self)
        self.mode_changed.emit(mode)

    def switch_mode(self, mode: MouseMode):
        self.set_mode(mode)
        self._last_mode_before_key_hold = mode
        if self.isVisible():
            mode_name = mode.name
            if mode_name.startswith("ROI_"):
                mode_name = mode_name[4:]
            show_tooltip(
                f"Mouse mode:<br>{mode_name}", duration=2, behavior="until_move"
            )

    def setSmoothing(self, enabled: bool):
        """Enable/disable image smoothing."""
        for im in self._image_widgets:
            im.setSmoothingEnabled(enabled)

    def scene(self) -> QBaseGraphicsScene:
        return super().scene()

    def resize_event(self, old_size: Size, new_size: Size):
        """Process the resize event."""
        if (w_new := new_size.width) < 10 or (h_new := new_size.height) < 10:
            return
        if (w_old := old_size.width) == 0 or (h_old := old_size.height) == 0:
            ratio = 1.0
        else:
            ratio = math.sqrt(w_new / w_old * h_new / h_old)

        self.scale_and_update_handles(ratio)
        self._inform_scale()
        if not self._initialized:
            self.initialize()

    def showEvent(self, event):
        self.initialize()
        return super().showEvent(event)

    def initialize(self):
        if self._initialized:
            return
        if len(self._image_widgets) == 0:
            return
        first = self._image_widgets[0]
        rect = first.boundingRect()
        if (size := max(rect.width(), rect.height())) <= 0:
            return
        factor = 1 / size
        self.scale_and_update_handles(factor)
        self.centerOn(rect.center())
        self._initialized = True

    def _inform_scale(self):
        show_tooltip(
            f"Zoom: {self.transform().m11():.3%}", duration=0.4, behavior="until_move"
        )

    def wheelEvent(self, event: QtGui.QWheelEvent):
        dy = event.angleDelta().y()
        self._wheel_event(dy)
        return None  # NOTE: don't call super().wheelEvent(event)

    def _wheel_event(self, dy: int):
        factor = 1.1
        if dy > 0:
            zoom_factor = factor
        else:
            zoom_factor = 1 / factor
        self.scale_and_update_handles(zoom_factor)
        self._inform_scale()

    def scale_and_update_handles(self, factor: float):
        """Scale the view and update the selection handle sizes."""
        if len(self._image_widgets) > 0:
            image_size = self._image_widgets[0]._qimage.size()
            length = min(image_size.width(), image_size.height())
            new_scale = self.transform().m11() * factor
            if length * new_scale < 1 or new_scale > 1000:
                # too small or too large, do not zoom
                return
        if factor > 0:
            self.scale(factor, factor)
        self.update_handle_sizes()
        self.geometry_changed.emit(self.sceneRect())

    def update_handle_sizes(self):
        tr = self.transform()
        self._selection_handles.update_handle_size(tr.m11())

    def auto_range(self):
        scene_rect = self.sceneRect()
        self.fitInView(scene_rect, Qt.AspectRatioMode.KeepAspectRatio)
        return None

    def remove_current_item(self, remove_from_list: bool = False, reason: str = ""):
        """Remove current ROI item.

        Parameters
        ----------
        remove_from_list : bool, optional
            If True, remove the item from the list of ROIs, otherwise just remove it
            from the scene.
        reason : str, optional
            Only used for logging.
        """
        if self._current_roi_item is not None:
            if not self._is_rois_visible:
                self._current_roi_item.setVisible(False)
            if remove_from_list:
                self.remove_item(self._current_roi_item)
            else:
                if self._is_current_roi_item_not_registered:
                    self.scene().removeItem(self._current_roi_item)
            self._selection_handles.clear_handles()
            self._current_roi_item = None
            self.current_roi_updated.emit()
            _LOGGER.debug(
                "remove_current_item(remove_from_list=%r, reason=%r)",
                remove_from_list,
                reason,
            )

    def remove_item(self, item: QRoi):
        """Remove selected ROI item from the view."""
        self.scene().removeItem(item)
        if not (
            item is self._current_roi_item and self._is_current_roi_item_not_registered
        ):
            idx = self._roi_items.index(item)
            del self._roi_items[idx]
            self.roi_removed.emit(idx)
        self._qroi_labels.update()

    def select_item(
        self,
        item: QtW.QGraphicsItem | None,
        is_registered_roi: bool = False,
    ):
        """Select the item during selection mode.

        Parameters
        ----------
        item : QGraphicsItem or None
            The item to select. If None, deselect the current item.
        is_registered_roi : bool, optional
            If True, the `item` is a considered as a registered ROI. This parameter is
            ususally used to select ROIs from the ROI list widget.
        """
        if item is not self._current_roi_item:
            self.remove_current_item(reason="deselect")
        if item is not None:
            self._selection_handles.connect_roi(item)
            self._is_current_roi_item_not_registered = not is_registered_roi
        if isinstance(item, QRoi):
            self._current_roi_item = item
            self.current_roi_updated.emit()
            item.setVisible(True)
            _LOGGER.debug("Item selected: %r", type(item).__name__)

    def select_item_at(self, pos: QtCore.QPointF) -> QRoi | None:
        """Select the item at the given position."""
        item_clicked = None
        is_registered = False
        if self._current_roi_item and self._current_roi_item.contains(pos):
            item_clicked = self._current_roi_item
            is_registered = not self._is_current_roi_item_not_registered
        elif self._is_rois_visible:
            for item in reversed(self._roi_items):
                if item.contains(pos):
                    item_clicked = item
                    is_registered = True
                    break
        self.select_item(item_clicked, is_registered_roi=is_registered)
        return item_clicked

    def _pos_to_tuple(self, pos: QtCore.QPointF) -> tuple[float, float]:
        if self._stick_to_grid:
            return int(round(pos.x())), int(round(pos.y()))
        return pos.x(), pos.y()

    def keyPressEvent(self, event):
        _key = event.key()
        if event.modifiers() == Qt.KeyboardModifier.NoModifier:
            if _key == Qt.Key.Key_Up and (item := self._current_roi_item):
                item.translate(0, -1)
                return
            elif _key == Qt.Key.Key_Down and (item := self._current_roi_item):
                item.translate(0, 1)
                return
            elif _key == Qt.Key.Key_Left and (item := self._current_roi_item):
                item.translate(-1, 0)
                return
            elif _key == Qt.Key.Key_Right and (item := self._current_roi_item):
                item.translate(1, 0)
                return
            # The super-class keyPressEvent does not forward the arrow keys. For this
            # widget, they should be forwarded.
        if _key in (Qt.Key.Key_Up, Qt.Key.Key_Down, Qt.Key.Key_Left, Qt.Key.Key_Right):
            return QtW.QWidget.keyPressEvent(self, event)
        return super().keyPressEvent(event)

    def mousePressEvent(self, event: QtGui.QMouseEvent):
        # Store the position of the mouse when the button is pressed
        if isinstance(item_under_cursor := self.itemAt(event.pos()), QHandleRect):
            # prioritize the handle mouse event
            self.scene().setGrabSource(item_under_cursor)
            return super().mousePressEvent(event)
        self.scene().setGrabSource(self)
        if event.button() == Qt.MouseButton.RightButton:
            pass
        else:
            if event.button() == Qt.MouseButton.MiddleButton:
                self._set_pan_zoom_temporary()
            else:
                with suppress(Exception):
                    ins = current_instance()
                    if ins.keys.contains("Space"):
                        self._set_pan_zoom_temporary()
            self._mouse_event_handler.pressed(event)
        return super().mousePressEvent(event)

    def mouseMoveEvent(self, event: QtGui.QMouseEvent):
        # Move the image using the mouse
        if event.button() == Qt.MouseButton.NoButton:
            self.hovered.emit(self.mapToScene(event.pos()))
        elif event.buttons() & Qt.MouseButton.LeftButton:
            if (
                self._mouse_event_handler._pos_drag_start is None
                or self._mouse_event_handler._pos_drag_prev is None
                or self.scene().grabSource() is not self
            ):
                return super().mouseMoveEvent(event)

        self._mouse_event_handler.moved(event)
        return super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event: QtGui.QMouseEvent):
        if event.button() == Qt.MouseButton.RightButton:
            if item := self.select_item_at(self.mapToScene(event.pos())):
                menu = self._make_menu_for_roi(item)
                menu.exec(event.globalPos())
            else:
                menu = self._make_menu_for_view()
                menu.exec(event.globalPos())
        else:
            if event.button() == Qt.MouseButton.MiddleButton:
                self.set_mode(self._last_mode_before_key_hold)
            self._mouse_event_handler.released(event)
        self.scene().setGrabSource(None)
        return super().mouseReleaseEvent(event)

    def mouseDoubleClickEvent(self, event: QtGui.QMouseEvent):
        if event.button() == Qt.MouseButton.LeftButton:
            self._mouse_event_handler.double_clicked(event)
        return super().mouseDoubleClickEvent(event)

    def move_items_by(self, dx: int, dy: int):
        """Translate items."""
        self.verticalScrollBar().setValue(self.verticalScrollBar().value() - dy)
        self.horizontalScrollBar().setValue(self.horizontalScrollBar().value() - dx)
        self.geometry_changed.emit(self.sceneRect())

    def set_current_roi(self, item: QtW.QGraphicsItem):
        self._current_roi_item = item
        self._is_current_roi_item_not_registered = True
        # To avoid the automatic translation of the scene visible region during drawing
        # ROIs, reset the scene rect after adding the item
        rect = self.scene().sceneRect()
        self.scene().addItem(item)
        self.scene().setSceneRect(rect)
        self.current_roi_updated.emit()
        _LOGGER.info(f"Set current ROI item to {item}")

    def add_current_roi(self):
        """Register the current ROI item."""
        if item := self._current_roi_item:
            self._is_current_roi_item_not_registered = False
            if len(self._roi_items) > 0 and self._roi_items[-1] is item:
                # do not add the same item
                return
            _LOGGER.info(f"Added ROI item {item}")
            self._selection_handles.finish_drawing_polygon()
            self._roi_items.append(item)
            self._qroi_labels.update()
            self.roi_added.emit(item)

    def standard_key_press(self, _key: Qt.Key, shift: bool = False):
        if _key == Qt.Key.Key_Space:
            self._set_pan_zoom_temporary()
        elif _key == Qt.Key.Key_R:
            if shift:
                self.switch_mode(MouseMode.ROI_ROTATED_RECTANGLE)
            else:
                self.switch_mode(MouseMode.ROI_RECTANGLE)
        elif _key == Qt.Key.Key_E:
            if shift:
                self.switch_mode(MouseMode.ROI_ROTATED_ELLIPSE)
            else:
                self.switch_mode(MouseMode.ROI_ELLIPSE)
        elif _key == Qt.Key.Key_P:
            # switch similar modes in turn
            if shift:
                self.switch_mode(MouseMode.ROI_POINTS)
            else:
                self.switch_mode(MouseMode.ROI_POINT)
        elif _key == Qt.Key.Key_L:
            if shift:
                self.switch_mode(MouseMode.ROI_SEGMENTED_LINE)
            else:
                self.switch_mode(MouseMode.ROI_LINE)
        elif _key == Qt.Key.Key_C:
            self.switch_mode(MouseMode.ROI_CIRCLE)
        elif _key == Qt.Key.Key_G:
            self.switch_mode(MouseMode.ROI_POLYGON)
        elif _key == Qt.Key.Key_S:
            self.switch_mode(MouseMode.SELECT)
        elif _key == Qt.Key.Key_Z:
            self.switch_mode(MouseMode.PAN_ZOOM)
        elif _key == Qt.Key.Key_T:
            self.add_current_roi()
        elif _key in (Qt.Key.Key_Delete, Qt.Key.Key_Backspace):
            self.remove_current_item(remove_from_list=True, reason="delete key")
        elif _key == Qt.Key.Key_V:
            self.set_show_rois(not self._is_rois_visible)
        # arrow keys -> translate the view

    def standard_ctrl_key_press(self, key: Qt.Key):
        if key == Qt.Key.Key_A:
            self.select_image()
        elif key == Qt.Key.Key_X:
            if self._current_roi_item is not None:
                item = self._current_roi_item.copy()
                self.remove_current_item(remove_from_list=True, reason="Ctrl+X")
                set_clipboard(text=str(item), internal_data=item)
        elif key == Qt.Key.Key_C:
            self._copy_current_roi()
        elif key == Qt.Key.Key_V:
            self._paste_roi()
        elif key == Qt.Key.Key_D:  # duplicate ROI
            if self._current_roi_item is not None:
                self._copy_current_roi()
                self._paste_roi()
        elif key == Qt.Key.Key_Up:
            self._wheel_event(120)
        elif key == Qt.Key.Key_Down:
            self._wheel_event(-120)

    def _copy_current_roi(self):
        if self._current_roi_item is not None:
            if self._is_current_roi_item_not_registered:
                self.add_current_roi()
            item = self._current_roi_item.copy()
            set_clipboard(text=str(item), internal_data=item)

    def _paste_roi(self) -> None:
        model = get_clipboard()
        item = model.internal_data
        if not isinstance(item, QRoi):
            return
        sx, sy = self.transform().m11(), self.transform().m22()
        delta = QtCore.QPointF(4 / sx, 4 / sy)
        dx, dy = self._pos_to_tuple(delta)
        item.translate(max(dx, 1), max(dy, 1))
        self.set_current_roi(item)
        self.add_current_roi()
        self._selection_handles.connect_roi(item)
        self.update_handle_sizes()
        set_clipboard(model.with_internal_data(item.copy()))  # needed for Ctrl+V x2

    def _make_menu_for_roi(self, roi: QRoi) -> QtW.QMenu:
        menu = QtW.QMenu(self)
        action = menu.addAction("Copy ROI")
        action.triggered.connect(
            lambda: set_clipboard(text=str(roi), internal_data=roi.copy())
        )
        action = menu.addAction("Delete ROI")
        action.triggered.connect(lambda: self.remove_item(roi))
        return menu

    def _make_menu_for_view(self):
        menu = QtW.QMenu(self)
        action = menu.addAction("Copy view to clipboard")
        action.triggered.connect(
            lambda: QtW.QApplication.clipboard().setImage(self.grab().toImage())
        )
        return menu

    def _set_pan_zoom_temporary(self):
        if (mode := self.mode()) is not MouseMode.PAN_ZOOM:
            self._last_mode_before_key_hold = mode
        self.set_mode(MouseMode.PAN_ZOOM)
